BeginPackage["QCD`"];
Unprotect["QCD`*"];

alphaS::usage =
"alphaS[mu] gives the strong coupling constant alpha_s at the scale mu."<>
"The value is obtained via evolution from alpha_s(mZ) = alphaSmZ"<>
" with 5 active flavours."<>
" alphaS[mu, {alphaS0, mu0, nf}] gives the strong coupling constant alpha_s"<>
" at the scale mu obtained via evolution from alpha_s(mu0) = alphaS0"<>
" with nf active flavours."<>
" Decoupling can be performed with the DecoupleDown and DecoupleUp Options.";

MSbarMass::usage =
"MSbarMass[mu, {m0, alphaS0, mu0, nf}] gives the MSbarMass m"<>
" at the scale mu obtained via evolution with nf active flavours"<>
" from m0 = m(mu0), using alpha_s(mu0) = alphaS0."<>
" Decoupling can be performed with the DecoupleDown and DecoupleUp Options.";

RunningOrder::usage =
"Option for alphaS."<>
" Specifies the order of the beta function that should be used for the"<>
" running of the strong coupling";
DecoupleDown::usage = "Specified a list of thresholds that should be removed"<>
" from the evolution of alphaS or MSbarMass."<>
" A threshold is a triple {mu, m, scheme},"<>
" where the first element specifies the decoupling scale,"<>
" the second element the mass of the decoupled quark,"<>
" and the third element the decoupling scheme \"OS\", \"MSbar\", or \"SI\"."<>
" For example, alphaS[mu, {alphaSmZ, mZ, 5}, DecoupleDown -> {{10, 5, \"OS\"}}]"<>
" gives alphaS(mu) in the four-flavour theory with on-shell decoupling"<>
" of the bottom quark with a mass of 5 GeV at the scale 10 GeV.";
DecoupleUp::usage = "Specified a list of thresholds that should be added"<>
" to the evolution of alphaS or MSbarMass."<>
" A threshold is a triple {mu, m, scheme},"<>
" where the first element specifies the decoupling scale,"<>
" the second element the mass of the decoupled quark,"<>
" and the third element the decoupling scheme \"OS\", \"MSbar\", or \"SI\"."<>
" For example, alphaS[mu, {alphaSmZ, mZ, 5}, DecoupleUp -> {{200, 175, \"OS\"}}]"<>
" gives alphaS(mu) in the six-flavour theory with on-shell decoupling"<>
" of the top quark with a mass of 175 GeV at the scale 200 GeV.";

alphaSmZ::usage = "Reference value for alpha_s(mZ) in the five-flavour theory";
mZ::usage = "Z boson mass.";

alphaS::decx = "Expect at most one of DecoupleUp or DecoupleDown to be set";

MSbarMass::decx = alphaS::decx;

Begin["`Private`"];
alphaSmZ= 0.1184;
mZ = 91.18;

RealQ = Internal`RealValuedNumericQ;

Options[alphaS] = {
   DecoupleUp -> {},
   DecoupleDown -> {},
   RunningOrder -> 4
};

Options[MSbarMass] = {
   DecoupleUp -> {},
   DecoupleDown -> {},
   RunningOrder -> 4
};

alphaS[mu_, opts:OptionsPattern[]] := alphaS[mu, {alphaSmZ, mZ, 5}, opts];

schemeToInt[scheme_String] := Switch[
   scheme,
   "Pole", 0,
   "OS", 0,
   "MSbar", 1,
   "MS", 1,
   "scale invariant", 2,
   "SI", 2
];

thrList[{a_?RealQ, b_?RealQ}] := {{a, b}, {schemeToInt["OS"]}};

thrList[{a_?RealQ, b_?RealQ, c_String}] := {{a, b}, {schemeToInt[c]}};

thrList[{thr__List}] := Module[
   {a, b, c, tmp},
   tmp = Replace[
      {thr},
      {
         {a_, b_} -> {a, b, schemeToInt["OS"]},
         {a_, b_, c_String} :> {a, b, schemeToInt[c]}
      },
      1
   ];
   Return[{Flatten[Most /@ tmp], Last /@ tmp}];
];

alphaS[
   mu_?RealQ, {as0_?RealQ, mu0_?RealQ, nf_Integer},
   OptionsPattern[]
] := Which[
   OptionValue[DecoupleUp] != {} && OptionValue[DecoupleDown] != {},
   Message[alphaS::decx],
   OptionValue[DecoupleDown] != {},
   asDecDown[
      mu, as0, mu0, nf, OptionValue[RunningOrder],
      Sequence @@ thrList[OptionValue[DecoupleDown]]
   ],
   OptionValue[DecoupleUp] != {},
   asDecUp[
      mu, as0, mu0, nf, OptionValue[RunningOrder],
      Sequence @@ thrList[OptionValue[DecoupleUp]]
   ],
   True,
   as[mu, as0, mu0, nf, OptionValue[RunningOrder]]
];

MSbarMass[
   mu_?RealQ, {m0_?RealQ, as0_?RealQ, mu0_?RealQ, nf_Integer},
   OptionsPattern[]
] := Which[
   OptionValue[DecoupleUp] != {} && OptionValue[DecoupleDown] != {},
   Message[alphaS::decx],
   OptionValue[DecoupleDown] != {},
   mDecDown[
      mu, m0, as0, mu0, nf, OptionValue[RunningOrder],
      Sequence @@ thrList[OptionValue[DecoupleDown]]
   ],
   OptionValue[DecoupleUp] != {},
   mDecUp[
      mu, m0, as0, mu0, nf, OptionValue[RunningOrder],
      Sequence @@ thrList[OptionValue[DecoupleUp]]
   ],
   True,
   m[mu, m0, as0, mu0, nf, OptionValue[RunningOrder]]
];

asDecUp = LibraryFunctionLoad[
   "libQCD_mathematica",
   "alpha_s_dec_up", {
   Real, Real, Real, Integer, Integer, {Real, 1}, {Integer, 1}
 },
 Real
];

asDecDown = LibraryFunctionLoad[
   "libQCD_mathematica",
   "alpha_s_dec_down", {
   Real, Real, Real, Integer, Integer, {Real, 1}, {Integer, 1}
 },
 Real
];

as = LibraryFunctionLoad[
   "libQCD_mathematica",
   "alpha_s", {
   Real, Real, Real, Integer, Integer
 },
 Real
];

mDecUp = LibraryFunctionLoad[
   "libQCD_mathematica",
   "m_dec_up", {
   Real, Real, Real, Real, Integer, Integer, {Real, 1}, {Integer, 1}
 },
 Real
];

mDecDown = LibraryFunctionLoad[
   "libQCD_mathematica",
   "m_dec_down", {
   Real, Real, Real, Real, Integer, Integer, {Real, 1}, {Integer, 1}
 },
 Real
];

m = LibraryFunctionLoad[
   "libQCD_mathematica",
   "m", {
   Real, Real, Real, Real, Integer, Integer
 },
 Real
];

End[];
Protect["QCD`*"];
EndPackage[];
