/**
 * @file
 * @brief Running and decoupling of quark masses in the \f$\overline{\rm MS}\f$ scheme
 * @author  Andreas Maier <andreas.maier@durham.ac.uk>
 * @version 2.1.1
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QCD library.
 *
 * The QCD library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QCD library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QCD library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include "alpha_s.hpp"
#include "gamma_m.hpp"
#include "beta.hpp"

//!
namespace QCD{
  namespace detail{

    //std::min is not constexpr in c++11
    template<typename T>
    constexpr T const & min(T const & a, T const & b){
      return (a<b)?a:b;
    }
  }

  //! Value for \f$\overline{\rm MS}\f$ quark mass at a reference scale
  struct reference_mass{
    //! Quark mass at the reference scale %mu
    double m;
    //! Reference scale
    double mu;
    //! Number of active flavours
    int nf;
  };

  struct Parameters {
    double alpha_s;
    double m;
  };

  //!Implements the running and decoupling of a quark mass in the \f$\overline{\rm MS}\f$ scheme.
  class MSbar_mass{
    double m0_;
    double as0_;
    double mu0_;
    double abserr_, relerr_;
    int nf_;
    int running_order_;

    std::array<double, 2> evolve(double mu) const;

    double decouple_mass_up(double alpha_s, threshold const & thr, int nf) const;
    double decouple_mass_down(double alpha_s, threshold const & thr, int nf) const;

    MSbar_mass() = default;

  public:
    //! Highest running order
    static constexpr int max_running_order =
      detail::min(gamma_m_max_order, beta_max_order);

    //! Default (relative and absolute) running error
    static constexpr double default_error = 1e-6;

    /**
     * Construct an \f$\overline{\rm MS}\f$ mass
     *
     * @param m            Reference mass with the corresponding scale
     *                     and number of active quark flavours
     * @param alpha_s      Strong coupling constant
     * @param order        Running order between 0 and #max_running_order
     *
     * @see #strong_coupling
     */
    MSbar_mass(
        reference_mass m,
        strong_coupling alpha_s,
	int order = max_running_order
    );

    /**
     * Construct an \f$\overline{\rm MS}\f$ mass
     *
     * @param m            Reference mass with the corresponding scale
     *                     and number of active quark flavours
     * @param order        Running order between 0 and #max_running_order
     * @param alpha_s_mu   Value of the strong coupling constant at
     *                     the reference scale m.mu
     *
     * @see #strong_coupling
     */
    MSbar_mass(
        reference_mass m,
        double alpha_s_mu,
	int order = max_running_order
    );

    /**
     * Evolve the mass to the given scale
     *
     * @param scale    Target scale for the evolution
     * @returns        The value of the mass at the given scale
     */
    double operator()(double scale) const;

    //! Decouple down (remove) flavour in running of mass
    /**
     * @param thr    The decoupling threshold
     * @return       The mass with nf - 1 active flavours
     */
    MSbar_mass decouple_down(threshold thr) const;

    //! Remove (decouple down) flavour in running of mass
    /**
     * @param thr    The decoupling threshold
     * @return       The mass with nf - 1 active flavours
     */
    MSbar_mass remove(threshold thr) const;

    //! Decouple up (add) flavour in running
    /**
     * @param thr    The decoupling threshold
     * @return       The mass with nf + 1 active flavours
     */
    MSbar_mass decouple_up(threshold thr) const;

    //! Add (decouple up) flavour in running
    /**
     * @param thr    The decoupling threshold
     * @return       The mass with nf + 1 active flavours
     */
    MSbar_mass add(threshold thr) const;

    //TODO: fix implementation
    // /**
    //  * Compute the scale-invariant mass \f$m(m)\f$
    //  *
    //  * @returns     The scale-invariant mass
    //  *
    //  */
    // double to_scale_invariant() const;

    /**
     * Compute the scale-invariant mass \f$m(m)\f$ with a given precision
     *
     * @param ep    Absolute target precision for the result
     * @returns     The scale-invariant mass
     */
    double to_scale_invariant(double ep) const;

    /**
     * Compute the scale-invariant mass \f$m(m)\f$
     *
     * @param t As in \ref to_scale_invariant
     * @returns The scale-invariant mass
     * @see to_scale_invariant
     *
     */
    template<typename... T>
    inline double to_SI(T&&... t) const{
      return to_scale_invariant(std::forward<T>(t)...);
    }

    /**
     * Get the current order for the evolution of the mass
     *
     * @returns     The running order
     */
    int order() const;

    /**
     * Get the number of active quark flavours
     *
     * @returns     The number of flavours
     */
    int nf() const;

    /**
     * Compute both the running mass and alpha_s at the given scale
     *
     * @param scale    Target scale for the evolution
     * @returns        The values of alpha_s and the mass at the given scale
     */
    Parameters with_alpha_s(double scale) const;

    //! Set error bounds for evolution
    /**
     * @param err    The maximum absolute and relative error
     */
    MSbar_mass & error(double err);

    //! Set absolute error bound for evolution
    /**
     * @param err    The maximum absolute error
     */
    MSbar_mass & abs_error(double abs_err);

    //! Set relative error bound for evolution
    /**
     * @param err    The maximum relative error
     */
    MSbar_mass & rel_error(double rel_err);

  };

}
