/**
 * @file
 * @brief Running and decoupling of the strong coupling constant
 * @author  Andreas Maier <andreas.maier@durham.ac.uk>
 * @version 2.1.1
 *
 * @section LICENSE
 *
 * Copyright 2014, 2018 Andreas Maier
 *
 * This file is part of the QCD library.
 *
 * The QCD library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QCD library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QCD library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <cassert>

#include "threshold.hpp"

//! -
namespace QCD{

  //! Default value for the strong coupling at the Z scale
  constexpr double alpha_s_mz = 0.1184;
  //! Default value for the Z boson mass
  constexpr double mz = 91.18;

  //! Value for strong coupling constant at a reference scale
  struct reference_coupling{
    //! Value of alpha_s at the reference scale %mu
    double alpha_s = alpha_s_mz;
    //! Reference scale
    double mu = mz;
    //! Number of active flavours
    int nf = 5;

    reference_coupling() = default;
    explicit reference_coupling(double alpha_s_mz): alpha_s(alpha_s_mz) {}
    reference_coupling(double alpha_s_mu0, double mu0, int nf0):
      alpha_s(alpha_s_mu0), mu(mu0), nf(nf0) {}
  };

  //! Implements the running and decoupling of the strong coupling constant
  class strong_coupling{
    double alpha_s_mu0;
    double mu0;
    double abserr_, relerr_;
    int nf0;
    int running_order;

    double evolve(
        double alpha_s, double mu0, double mu, int nf,
        double abserr, double relerr
    ) const;
    double decouple_up(double alpha_s, threshold thr, int nf) const;
    double decouple_down(double alpha_s, threshold thr, int nf) const;
  public:
    //! Highest running order for the strong coupling
    static constexpr int max_running_order = 4;

    //! Default (relative and absolute) running error
    static constexpr double default_error = 1e-6;

    /**
     * Construct a new strong coupling
     *
     * @param alpha_s    Reference value with the corresponding scale     *                   and number of active flavours
     * @param order      Running order between 0 and #max_order
     *
     */
    explicit strong_coupling(
        reference_coupling alpha_s = reference_coupling{},
	int order = max_running_order
    );

    /**
     * Evolve the strong coupling to the given scale
     *
     * @param mu    Target scale for the evolution
     * @returns     The value of the strong coupling
     */
    double operator()(double mu) const;

    /**
     * Get the current order for the evolution of the coupling
     *
     * @returns     The running order
     */
    int order() const;

    /**
     * Get the number of active quark flavours
     *
     * @returns     The number of flavours
     */
    int nf() const;

    //! Decouple down (remove) flavour in alpha_s
    /**
     * @param thr    The decoupling threshold
     * @return       The strong coupling with nf - 1 active flavours
     */
    strong_coupling decouple_down(threshold thr) const;
    //! Remove (decouple down) flavour in alpha_s
    /**
     * @param thr    The decoupling threshold
     * @return       The strong coupling with nf - 1 active flavours
     */
    strong_coupling remove(threshold thr) const;

    //! Decouple up (add) flavour in alpha_s
    /**
     * @param thr    The decoupling threshold
     * @return       The strong coupling with nf + 1 active flavours
     */
    strong_coupling decouple_up(threshold thr) const;

    //! Add (decouple up) flavour in alpha_s
    /**
     * @param thr    The decoupling threshold
     * @return       The strong coupling with nf + 1 active flavours
     */
    strong_coupling add(threshold thr) const;

    //! Set error bounds for evolution
    /**
     * @param err    The maximum absolute and relative error
     */
    strong_coupling & error(double err);

    //! Set absolute error bound for evolution
    /**
     * @param err    The maximum absolute error
     */
    strong_coupling & abs_error(double abs_err);

    //! Set relative error bound for evolution
    /**
     * @param err    The maximum relative error
     */
    strong_coupling & rel_error(double rel_err);

  };

  //! Default strong coupling constant in five-flavour theory
  static const strong_coupling alpha_s_nf5{};
  //! Default strong coupling constant in four-flavour theory
  /**
   * %alpha_s_nf4 is obtained from %alpha_s_nf5 via decoupling
   * at the default %bottom_threshold
   */
  static const strong_coupling alpha_s_nf4 =
    alpha_s_nf5.decouple_down(bottom_threshold);
  //! Default strong coupling constant in six-flavour theory
  /**
   * %alpha_s_nf6 is obtained from %alpha_s_nf5 via decoupling
   * at the default %top_threshold
   */
  static const strong_coupling alpha_s_nf6 =
    alpha_s_nf5.decouple_up(top_threshold);

}
