/**
 * @file
 * @brief Mass anomalous dimension in the \f$\overline{\rm MS}\f$ scheme
 * @author  Andreas Maier <andreas.maier@durham.ac.uk>
 * @version 2.1.1
 *
 * @section LICENSE
 *
 * Copyright 2014 Andreas Maier
 *
 * This file is part of the QCD library.
 *
 * The QCD library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QCD library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QCD library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <array>

#if __cplusplus >= 201402L
#define QCD_CONSTEXPR_AFTER_CXX14 constexpr
#else
#define QCD_CONSTEXPR_AFTER_CXX14
#endif

namespace QCD{

  constexpr unsigned gamma_m_max_order = 4;

  namespace detail{

    namespace gamma_m{
      constexpr double pi = 3.141592653589793238462643383279;
      constexpr double zeta_3 = 1.202056903159594285399738161511449990765;
      constexpr double zeta_4 = pi*pi*pi*pi/90.;
      constexpr double zeta_5 = 1.036927755143369926331365486457034168057;
      constexpr double zeta_6 = 1.017343061984449139714517929790920527902;
      constexpr double zeta_7 = 1.008349277381922826839797549849796759600;

      constexpr auto max_order = gamma_m_max_order;

      // gamma_m = - sum gamma_m_ij nf^j as^(i+1)
      // as = alpha_s/pi
      //taken from Baikov, Chetyrkin, Kuehn arXiv:1402.6611
      constexpr double gamma_m_00 = 1;

      constexpr double gamma_m_10 = 101./24.;
      constexpr double gamma_m_11 = -5./36.;

      constexpr double gamma_m_20 = 1249./64.;
      constexpr double gamma_m_21 = -277./216. - 5./6.*zeta_3;
      constexpr double gamma_m_22 = -35./1296.;

      constexpr double gamma_m_30 =
	4603055./41472. + 530./27.*zeta_3 - 275./8.*zeta_5;
      constexpr double gamma_m_31 =
	-91723./6912. - 2137./144.*zeta_3 + 55./16.*zeta_4 + 575./72.*zeta_5;
      constexpr double gamma_m_32 =
	2621./31104. + 25./72.*zeta_3 - 5./24.*zeta_4;
      constexpr double gamma_m_33 = -83./15552. + zeta_3/108.;

      constexpr double gamma_m_40 =
	+ 99512327./165888. + 23201233./124416.*zeta_3 + 3025./32.*zeta_3*zeta_3
	- 349063./4608.*zeta_4 - 28969645./31104.*zeta_5 + 15125./64.*zeta_6
	+ 25795./64.*zeta_7
	;
      constexpr double gamma_m_41 =
	- 150736283./1492992. - 391813./2592.*zeta_3 - 2365./288.*zeta_3*zeta_3
	+ 1019371./13824.*zeta_4 + 12469045./62208.*zeta_5 - 39875./576.*zeta_6
	- 56875./864.*zeta_7
	;
      constexpr double gamma_m_42 =
	+ 660371./373248. + 251353./31104.*zeta_3 + 725./432.*zeta_3*zeta_3 -
	41575./6912.*zeta_4 - 33005./10368.*zeta_5 + 2875./864.*zeta_6
	;
      constexpr double gamma_m_43 =
	91865./1492992. + 803./5184.*zeta_3 + 7./144.*zeta_4 - 5./27.*zeta_5
	;
      constexpr double gamma_m_44 = -65./62208. - 5./3888.*zeta_3 + zeta_4/432.;

      constexpr auto max_nf_pow =
	std::array<unsigned,1u + max_order>{{0u, 1u, 2u, 3u, 4u}};

      inline QCD_CONSTEXPR_AFTER_CXX14
      unsigned coeffs_start_index(unsigned order){
	return (order == 0)?
	  0:
	  (1u + max_nf_pow[order-1] + coeffs_start_index(order-1))
	  ;
      }

      constexpr auto coeffs = std::array<double,15>{{
	  gamma_m_00,
	  gamma_m_10, gamma_m_11,
	  gamma_m_20, gamma_m_21, gamma_m_22,
	  gamma_m_30, gamma_m_31, gamma_m_32, gamma_m_33,
	  gamma_m_40, gamma_m_41, gamma_m_42, gamma_m_43, gamma_m_44
	}};

      inline QCD_CONSTEXPR_AFTER_CXX14
      unsigned gamma_m_max_nf_pow(unsigned order){
	// assert(order <= max_order);
	using namespace detail::gamma_m;
	return max_nf_pow[order];
      }

      inline QCD_CONSTEXPR_AFTER_CXX14
      double gamma_m_nf_coeff(unsigned order, unsigned nfpow){
	// assert(order <= max_order);
	// assert(nfpow <= max_nf_pow(order));
	using namespace detail::gamma_m;
	return( coeffs[coeffs_start_index(order) + nfpow] );
      }

      inline QCD_CONSTEXPR_AFTER_CXX14
      double helper(unsigned order, int nf, unsigned nf_pow){
	//assert(nf_pow <= max_nf_pow[order])
	return
	  (nf_pow == max_nf_pow[order])?
	  gamma_m_nf_coeff(order, nf_pow):
	  gamma_m_nf_coeff(order, nf_pow) + nf*helper(order, nf, nf_pow + 1)
	  ;
      }

      inline QCD_CONSTEXPR_AFTER_CXX14
      double gamma_m(unsigned order, int nf){
	// assert(order <= max_order);
	using namespace detail::gamma_m;
	return helper(order, nf, 0);
      }

      inline QCD_CONSTEXPR_AFTER_CXX14
      double helper(
	  unsigned order,
	  int nf,
	  double alpha_s,
	  unsigned cur_order
      ){
	//assert(order <= max_order)
	return (cur_order == order)?
	  gamma_m(order, nf):(
	      + gamma_m(cur_order, nf)
	      + alpha_s/pi * helper(order, nf, alpha_s, cur_order + 1)
	  );
      }
    }
  }

  //! The QCD mass anomalous dimension
  /**
   * @param order   Perturbative order of the mass anomalous dimension
   * @param nf      Number of active flavours
   * @param alpha_s Value of the strong coupling constant
   * @returns       The value of the mass anomalous dimension
   *
   * The mass anomalous dimension is defined via the renormalisation group equation
   * \f$\gamma^m = \frac{d\ln m(\mu)}{d\ln \mu^2} \f$, where
   * \f$m(\mu)\f$ is the running mass at the scale \f$\mu\f$.
   *
   */
  inline QCD_CONSTEXPR_AFTER_CXX14
  double gamma_m(unsigned order, int nf, double alpha_s){
    // assert(order <= max_order);
    using namespace detail::gamma_m;
    return - alpha_s/pi*helper(order, nf, alpha_s, 0);
  }

  //! Perturbative coefficients of the QCD mass anomalous dimension
  /**
   * @param order   Perturbative order of the coefficient
   * @param nf      Number of active flavours
   * @returns       The value of the coefficient of the mass anomalous dimension
   *
   * The coefficients of the mass anomalous dimension are given by
   * \f$\gamma^m = - \sum_{i=0}^\infty \gamma^m_i (\frac{\alpha_s}{\pi})^{i+1}\f$.
   * For example, gamma_m(1, 5) returns the value of \f$\gamma^m_0\f$
   * for 5 active flavours.
   *
   */
  inline QCD_CONSTEXPR_AFTER_CXX14
  double gamma_m(unsigned order, int nf){
    return detail::gamma_m::gamma_m(order, nf);
  }

  //! Highest power of number of flavours in appearing in the mass anomalous dimension
  /**
   * @param order   Perturbative order of the mass anomalous dimension
   * @returns       The highest power of the number of active flavours
   *                in the formula for the mass anomalous dimension up to the given order
   */
  inline QCD_CONSTEXPR_AFTER_CXX14
  unsigned gamma_m_max_nf_pow(unsigned order){
    return detail::gamma_m::gamma_m_max_nf_pow(order);
  }
  //! Coefficient of some power of number of flavours in the mass anomalous dimension
  /**
   * @param order   Perturbative order
   * @param nfpow   Power of number of active flavour
   * @returns       The coefficient of \f$n_f^{\rm nfpow}\f$ in the formula for \f$\gamma^m_{\rm order}\f$
   */
  inline QCD_CONSTEXPR_AFTER_CXX14
  double gamma_m_nf_coeff(unsigned order, unsigned nfpow){
    return detail::gamma_m::gamma_m_nf_coeff(order, nfpow);
  }


  // the following functions are just defined for safety, e.g.
  // to prevent unintended implicit conversion from negative ints
  // please avoid using them

  inline QCD_CONSTEXPR_AFTER_CXX14
  double gamma_m_nf_coeff(int order, unsigned nfpow){
    // assert(order >= 0);
    return gamma_m_nf_coeff(static_cast<unsigned>(order), nfpow);
  }

  inline QCD_CONSTEXPR_AFTER_CXX14
  double gamma_m_nf_coeff(unsigned order, int nfpow){
    // assert(nfpow >= 0);
    return gamma_m_nf_coeff(order, static_cast<unsigned>(nfpow));
  }

  inline QCD_CONSTEXPR_AFTER_CXX14
  double gamma_m_nf_coeff(int order, int nfpow){
    // assert(order >= 0);
    // assert(nfpow >= 0);
    return
      gamma_m_nf_coeff(static_cast<unsigned>(order), static_cast<unsigned>(nfpow));
  }

}

#undef QCD_CONSTEXPR_AFTER_CXX14
