/**
 * @file
 * @brief Decoupling constant for the \f$\overline{\rm MS}\f$ mass.
 * @author  Andreas Maier <andreas.maier@durham.ac.uk>
 * @version 2.1.1
 *
 * @section LICENSE
 *
 * Copyright 2014, 2021 Andreas Maier
 *
 * This file is part of the QCD library.
 *
 * The QCD library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QCD library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QCD library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <array>
#include <cstddef>

#include "scheme.hpp"

#if __cplusplus >= 201402L
#define QCD_CONSTEXPR_AFTER_CXX14 constexpr
#else
#define QCD_CONSTEXPR_AFTER_CXX14
#endif

//! -
namespace QCD{

  // decoupling coefficients for MSbar mass from
  // Chetyrkin, Kniehl, Steinhauser hep-ph/9708255
  // Chetyrkin, Kuehn, Steinhauser  hep-ph/0004189
  // Liu, Steinhauser arXiv:1502.04719
  // m^(nf-1) = zeta_m * m^(nf)
  // m^(nf+1) = zeta_m_inv * m^(nf)
  // lm = log(mu^2/m^2)

  constexpr unsigned zeta_m_max_order = 4;

  namespace detail{

    namespace zeta_m{
      constexpr double pi = 3.141592653589793238462643383279;

      constexpr double zeta_2 = pi*pi/6.;
      constexpr double zeta_3 = 1.202056903159594285399738161511449990765;
      constexpr double zeta_4 = pi*pi*pi*pi/90.;
      constexpr double zeta_5 = 1.036927755143369926331365486457034168057;
      constexpr double zeta_6 = pi*pi*pi*pi*pi*pi/945.;
      constexpr double log2 = 0.6931471805599453094172321214581765680755;
      constexpr double B4 = -1.762800087073770864061897634679818807215;
      // c4 = 24*Li_4(1/2) + log^4(2) - pi^2*log^2(2)
      constexpr double c4 = 7.908451398072941071188189589256658841231;
      // c5 = -360*Li_5(1/2) + 3*log^5(2) - 5*pi^2*log^3(2)
      constexpr double c5 = -198.9783082886512628182933723717652048993;

      constexpr unsigned max_nf_pow(unsigned order){
	return (order > 2u)?(order-2u):0u;
      }

      constexpr unsigned max_log_pow(unsigned order, unsigned){
	return (order > 1u)?order:0u;
      }

      // coefficients of zeta_m
      constexpr std::size_t num_coeff = 28;
      constexpr auto data_OS = std::array<double, num_coeff>{{
      	  + 1., //alpha_s^0
      	  + 0., //alpha_s^1
	  89./432., -5./36., 1./12., // alpha_s^2 log^0...2
	  6157./11664. - B4/36. - 343./864.*zeta_3 + 5./4.*zeta_4, // alpha_s^3 nf^0 log^0
      	  439./2592. -5./6.*zeta_3, 319./432., 31./216., //alpha_s^3 nf^0 log^1...3
	  1327./11664. - 2./27.*zeta_3, -53./432., 0., -1./108., //alpha_s^3 nf^1 log^0...3
          133533622679./3292047360. - 2006969./4354560.*c4 + 13./972.*c5
          - 35./54.*zeta_2 - 5./27.*log2*zeta_2 - 416579531./40642560.*zeta_3
          + 1061./576.*zeta_3*zeta_3 + 16350665./580608.*zeta_4 - 1597./216.*log2*zeta_4
          - 61997./1728.*zeta_5 - 3935./432.*zeta_6, // alpha_s^4 nf^0 log^0
          -412745./124416. - 11./72.*c4 + 7./9.*zeta_2 + 2./9.*log2*zeta_2
          - 143231./9216.*zeta_3 + 1483./96.*zeta_4 + 575./72.*zeta_5, // alpha_s^4 nf^0 log^1
          28663./2592. - 155./48.*zeta_3, 6035./2592., 1039./3456., // alpha_s^4 nf^0 log^2...4
          -2615471./746496. + 49./2592.*c4 - c5/810 + 5./54.*zeta_2 - 1055./1728.*zeta_3
          - 721./3456.*zeta_4 + 49./72.*log2*zeta_4 + 497./288.*zeta_5, // alpha_s^4 nf^1 log^0
          34739./93312. + c4/108 - zeta_2/9. + 439./1728.*zeta_3 - 163./144.*zeta_4, // alpha_s^4 nf^1 log^1
          -10097./10368., -47./288., -1./27., // alpha_s^4 nf^1 log^2...4
          17671./124416. - 5./864.*zeta_3 - 7./96.*zeta_4, // alpha_s^4 nf^2 log^0
          -3401./46656. + 7./108.*zeta_3, 31./1296., 0., 1./864. // alpha_s^4 nf^2 log^2...4
      	}};

      constexpr auto data_MS = std::array<double, num_coeff>{{
      	  1.,
      	  0.,
      	  89./432., -5./36., 1./12.,
      	  10477./11664. - B4/36. - 343./864.*zeta_3 + 5./4.*zeta_4,
      	  7./2592. - 5./6.*zeta_3, 175./432., 31./216.,
      	  1327./11664. - 2./27.*zeta_3, -53./432., 0., -1./108.,
          142408730039./3292047360. - 2006969./4354560.*c4 + 13./972.*c5
          - 328144331./40642560.*zeta_3 + 1061./576.*zeta_3*zeta_3 + 16350665./580608.*zeta_4
          - 1597./216.*log2*zeta_4 - 61997./1728.*zeta_5 - 3935./432.*zeta_6,
          -1012697./124416. - 11./72.*c4 - 14151./1024.*zeta_3
          + 1483./96.*zeta_4 + 575./72.*zeta_5,
          14809./2592. - 155./48.*zeta_3, 2615./2592., 1039./3456.,
          -2473487./746496. + 49./2592.*c4 - c5/810 - 1055./1728.*zeta_3
          - 721./3456.*zeta_4 + 49./72.*log2*zeta_4 + 497./288.*zeta_5,
          63611./93312. + c4/108 + 439./1728.*zeta_3 - 163./144.*zeta_4,
          -8321./10368., -23./288., -1./27.,
          17671./124416. - 5./864.*zeta_3 - 7./96.*zeta_4,
          -3401./46656. + 7./108.*zeta_3, 31./1296., 0., 1./864.
      	}};

      constexpr auto data_SI = std::array<double, num_coeff>{{
      	  1.,
      	  0.,
      	  89./432., -5./36., 1./12.,
      	  10477./11664. - B4/36. - 343./864.*zeta_3 + 5./4.*zeta_4,
      	  -713./2592. - 5./6.*zeta_3, 319./432., 31./216.,
      	  1327./11664. - 2./27.*zeta_3, -53./432., 0., -1./108.,
          142408730039./3292047360. - 2006969./4354560.*c4 + 13./972.*c5
          - 328144331./40642560.*zeta_3 + 1061./576.*zeta_3*zeta_3 + 16350665./580608.*zeta_4
          - 1597./216.*log2*zeta_4 - 61997./1728.*zeta_5 - 3935./432.*zeta_6,
          -1157465./124416. - 11./72.*c4 - 47573./3072.*zeta_3
          + 1483./96.*zeta_4 + 575./72.*zeta_5,
          22519./2592. - 155./48.*zeta_3, 6035./2592., 1039./3456.,
          -2473487./746496. + 49./2592.*c4 - c5/810 - 1055./1728.*zeta_3
          - 721./3456.*zeta_4 + 49./72.*log2*zeta_4 + 497./288.*zeta_5,
          44315./93312. + c4/108 + 439./1728.*zeta_3 - 163./144.*zeta_4,
          -8561./10368., -47./288., -1./27.,
          17671./124416. - 5./864.*zeta_3 - 7./96.*zeta_4,
          -3401./46656. + 7./108.*zeta_3, 31./1296., 0, 1./864.
      	}};


      template<scheme s>
      QCD_CONSTEXPR_AFTER_CXX14
      double coeffs(std::size_t);

      template<>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double coeffs<scheme::pole>(std::size_t i){
	return data_OS[i];
      }

      template<>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double coeffs<scheme::MSbar>(std::size_t i){
	return data_MS[i];
      }

      template<>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double coeffs<scheme::scale_invariant>(std::size_t i){
	return data_SI[i];
      }

      /* coefficients of zeta_m^-1 */
      constexpr auto data_OS_inv = std::array<double, num_coeff>{{
      	  1., //alpha_s^0
      	  0., //alpha_s^1
      	  -89./432., 5./36., -1./12., //alpha_s^2 log^0...2
      	  -1871/2916. + B4/36. + 407./864.*zeta_3 - 5./4.*zeta_4, //alpha_s^3 nf^0 log^0
      	  -299./2592. + 5./6.*zeta_3, -299./432., -35./216., //alpha_s^3 nf^0 log^1...3
      	  -1327./11664. + 2./27.*zeta_3, 53./432., 0., 1./108., //alpha_s^3 nf^1 log^0...3
          -122722873709./3292047360. + 1924649./4354560.*c4 - 59./4860.*c5
          + 5./9.*zeta_2 + 5./27.*log2*zeta_2 + 441628331./40642560.*zeta_3
          - 1061./576.*zeta_3*zeta_3 - 16187201./580608.*zeta_4 + 725./108.*log2*zeta_4
          + 59015./1728.*zeta_5 + 3935./432.*zeta_6, //alpha_s^4 nf^0 log^0
          99319./41472. + 11./72.*c4 - 2./3.*zeta_2 - 2./9.*log2*zeta_2
          + 142463./9216.*zeta_3 - 481./32.*zeta_4 - 575./72.*zeta_5, //alpha_s^4 nf^0 log^1
          -34297./3456. + 175./48.*zeta_3, -6941./2592., -1147./3456., //alpha_s^4 nf^0 log^2...4
          2403419./746496. - 49./2592.*c4 + c5/810 - 5./54.*zeta_2 + 1075./1728.*zeta_3
          + 1225./3456.*zeta_4 - 49./72.*log2*zeta_4 - 497./288.*zeta_5, //alpha_s^4 nf^1 log^0
          -26443./93312. - c4/108 + zeta_2/9. - 599./1728.*zeta_3 + 163./144.*zeta_4, //alpha_s^4 nf^1 log^1
          10237./10368., 47./288., 17./432., //alpha_s^4 nf^1 log^2...4
          -17671./124416. + 5./864.*zeta_3 + 7./96.*zeta_4, //alpha_s^4 nf^2 log^0
          3401./46656. - 7./108.*zeta_3, -31./1296., 0., -1./864. //alpha_s^4 nf^2 log^1...4
      	}};


      constexpr auto data_MS_inv = std::array<double, num_coeff>{{
      	  1., //alpha_s^0
      	  0., //alpha_s^1
      	  -89./432., 5./36., -1./12., //alpha_s^2 log^0...2
      	  -2951./2916. + B4/36. + 407./864.*zeta_3 - 5./4.*zeta_4, //alpha_s^3 nf^0 log^0
      	  133./2592. + 5./6.*zeta_3, -155./432., -35./216., //alpha_s^3 nf^0 log^1...3
      	  -1327./11664. + 2./27.*zeta_3, 53./432., 0., 1./108., //alpha_s^3 nf^1 log^0...3
          -131621265869./3292047360. + 1924649./4354560.*c4 - 59./4860.*c5
          + 353193131./40642560.*zeta_3 - 1061./576.*zeta_3*zeta_3 - 16187201./580608.*zeta_4
          + 725./108.*log2*zeta_4 + 59015./1728.*zeta_5 + 3935./432.*zeta_6, //alpha_s^4 nf^0 log^0
          279367./41472. + 11./72.*c4 + 42197./3072.*zeta_3
          - 481./32.*zeta_4 - 575./72.*zeta_5, //alpha_s^4 nf^0 log^1
          -16193./3456. + 175./48.*zeta_3, -3161./2592., -1147./3456., //alpha_s^4 nf^0 log^2...4
          2261435./746496. - 49./2592.*c4 + c5/810 + 1075./1728.*zeta_3
          + 1225./3456.*zeta_4 - 49./72.*log2*zeta_4 - 497./288.*zeta_5, //alpha_s^4 nf^1 log^0
          -55315./93312. - c4/108 - 599./1728.*zeta_3 + 163./144.*zeta_4, //alpha_s^4 nf^1 log^1
          8461./10368., 23./288., 17./432., //alpha_s^4 nf^1 log^2...4
          -17671./124416. + 5./864.*zeta_3 + 7./96.*zeta_4, //alpha_s^4 nf^2 log^0
          3401./46656. - 7./108.*zeta_3, -31./1296., 0, -1./864. //alpha_s^4 nf^2 log^1...4
      	}};

      constexpr auto data_SI_inv = std::array<double, num_coeff>{{
      	  1., //alpha_s^0
      	  0., //alpha_s^1
      	  -89./432., 5./36., -1./12., //alpha_s^2 log^0...2
      	  -2951./2916. + B4/36. + 407./864.*zeta_3 - 5./4.*zeta_4, //alpha_s^3 nf^0 log^0
      	  853./2592. + 5./6.*zeta_3, -299./432., -35./216., //alpha_s^3 nf^0 log^1...3
      	  -1327./11664. + 2./27.*zeta_3, 53./432., 0., 1./108., //alpha_s^3 nf^1 log^0...3
          -131621265869./3292047360. + 1924649./4354560.*c4 - 59./4860.*c5
          + 353193131./40642560.*zeta_3 - 1061./576.*zeta_3*zeta_3 - 16187201./580608.*zeta_4
          + 725./108.*log2*zeta_4 + 59015./1728.*zeta_5 + 3935./432.*zeta_6, //alpha_s^4 nf^0 log^0
          330503./41472. + 11./72.*c4 + 47317./3072.*zeta_3
          - 481./32.*zeta_4 - 575./72.*zeta_5, //alpha_s^4 nf^0 log^1
          -8531./1152. + 175./48.*zeta_3, -6941./2592., -1147./3456., //alpha_s^4 nf^0 log^2...4
          2261435./746496. - 49./2592.*c4 + c5/810 + 1075./1728.*zeta_3
          + 1225./3456.*zeta_4 - 49./72.*log2*zeta_4 - 497./288.*zeta_5, //alpha_s^4 nf^1 log^0
          -36019./93312. - c4/108 - 599./1728.*zeta_3 + 163./144.*zeta_4, //alpha_s^4 nf^1 log^1
          8701./10368., 47./288., 17./432., //alpha_s^4 nf^1 log^2...4
          -17671./124416. + 5./864.*zeta_3 + 7./96.*zeta_4, //alpha_s^4 nf^2 log^0
          3401./46656. - 7./108.*zeta_3, -31./1296., 0, -1./864. //alpha_s^4 nf^2 log^1...4
      	}};

      template<scheme s>
      QCD_CONSTEXPR_AFTER_CXX14
      double coeffs_inv(std::size_t);

      template<>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double coeffs_inv<scheme::pole>(std::size_t i){
	return data_OS_inv[i];
      }

      template<>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double coeffs_inv<scheme::MSbar>(std::size_t i){
	return data_MS_inv[i];
      }

      template<>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double coeffs_inv<scheme::scale_invariant>(std::size_t i){
	return data_SI_inv[i];
      }


      // get index in coeffs array for a given order and power of nf, lm
      constexpr unsigned get_index(
	  unsigned order,
	  unsigned nf_pow
      );

      constexpr unsigned get_index(
	  unsigned order
      ){
	return (order == 0u)?
	  0:(
              + 1u
              + max_log_pow(order - 1, max_nf_pow(order - 1))
              + get_index(order - 1, max_nf_pow(order - 1))
	  );
      }

      constexpr unsigned get_index(
	  unsigned order,
	  unsigned nf_pow
      ){
	return (nf_pow == 0u)?
	  get_index(order):(
              1u + max_log_pow(order, nf_pow - 1) + get_index(order, nf_pow - 1)
	  );
      }

      constexpr unsigned get_index(
	  unsigned order,
	  unsigned nf_pow,
	  unsigned log_pow
      ){
	return log_pow + get_index(order, nf_pow);
      }
    }
  }

  template<scheme s>
  inline QCD_CONSTEXPR_AFTER_CXX14
  double zeta_m_log_coeff(
      unsigned order,
      unsigned nf_pow,
      unsigned log_pow
  ){
    using namespace detail::zeta_m;
    return coeffs<s>(get_index(order, nf_pow, log_pow));
  }

  template<scheme s>
  inline QCD_CONSTEXPR_AFTER_CXX14
  double zeta_m_inv_log_coeff(
      unsigned order,
      unsigned nf_pow,
      unsigned log_pow
  ){
    using namespace detail::zeta_m;
    return coeffs_inv<s>(get_index(order, nf_pow, log_pow));
  }


  namespace detail{
    namespace zeta_m{

      template<scheme s>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double nf_coeff_helper(
	  unsigned order,
	  unsigned nf_pow,
	  double lm,
	  unsigned cur_log_pow
      ){
	return (cur_log_pow == max_log_pow(order, nf_pow))?
	  zeta_m_log_coeff<s>(order, nf_pow, cur_log_pow):(
	      + zeta_m_log_coeff<s>(order, nf_pow, cur_log_pow)
	      + lm * nf_coeff_helper<s>(order, nf_pow, lm, cur_log_pow + 1)
	  );
      }

      template<scheme s>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double nf_coeff_helper_inv(
	  unsigned order,
	  unsigned nf_pow,
	  double lm,
	  unsigned cur_log_pow
      ){
	return (cur_log_pow == max_log_pow(order, nf_pow))?
	  zeta_m_inv_log_coeff<s>(order, nf_pow, cur_log_pow):(
	      + zeta_m_inv_log_coeff<s>(order, nf_pow, cur_log_pow)
	      + lm * nf_coeff_helper_inv<s>(order, nf_pow, lm, cur_log_pow + 1)
	  );
      }
    }
  }

  template<scheme s>
  inline QCD_CONSTEXPR_AFTER_CXX14
  double zeta_m_nf_coeff(
      unsigned order,
      unsigned nf_pow,
      double lm
  ){
    using namespace detail::zeta_m;
    return nf_coeff_helper<s>(order, nf_pow, lm, 0);
  }

  template<scheme s>
  inline QCD_CONSTEXPR_AFTER_CXX14
  double zeta_m_inv_nf_coeff(
      unsigned order,
      unsigned nf_pow,
      double lm
  ){
    using namespace detail::zeta_m;
    return nf_coeff_helper_inv<s>(order, nf_pow, lm, 0);
  }

  namespace detail{
    namespace zeta_m{
      template<scheme s>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double helper(
	  unsigned order,
	  unsigned nf,
	  double lm,
	  unsigned cur_nf_pow
      ){
	return (cur_nf_pow == max_nf_pow(order))?
	  zeta_m_nf_coeff<s>(order, cur_nf_pow, lm):(
	      + zeta_m_nf_coeff<s>(order, cur_nf_pow, lm)
	      + nf * helper<s>(order, nf, lm, cur_nf_pow + 1)
	  );
      }

      template<scheme s>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double helper_inv(
	  unsigned order,
	  unsigned nf,
	  double lm,
	  unsigned cur_nf_pow
      ){
	return (cur_nf_pow == max_nf_pow(order))?
	  zeta_m_inv_nf_coeff<s>(order, cur_nf_pow, lm):(
	      + zeta_m_inv_nf_coeff<s>(order, cur_nf_pow, lm)
	      + nf * helper_inv<s>(order, nf, lm, cur_nf_pow + 1)
	  );
      }

      template<scheme s>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double zeta_m(
	  unsigned order,
	  unsigned nf,
	  double lm
      ){
	return helper<s>(order, nf, lm, 0);
      }

      template<scheme s>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double zeta_m_inv(
	  unsigned order,
	  unsigned nf,
	  double lm
      ){
	return helper_inv<s>(order, nf, lm, 0);
      }

      template<scheme s>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double helper(
	  unsigned order,
	  unsigned nf,
	  double lm,
	  double alpha_s,
	  unsigned cur_order
      ){
	return (cur_order == order)?
	  zeta_m<s>(order, nf, lm):(
              + zeta_m<s>(cur_order, nf, lm)
              + alpha_s/pi * helper<s>(order, nf, lm, alpha_s, cur_order + 1)
	  );
      }

      template<scheme s>
      inline QCD_CONSTEXPR_AFTER_CXX14
      double helper_inv(
	  unsigned order,
	  unsigned nf,
	  double lm,
	  double alpha_s,
	  unsigned cur_order
      ){
	return (cur_order == order)?
	  zeta_m_inv<s>(order, nf, lm):(
              + zeta_m_inv<s>(cur_order, nf, lm)
              + alpha_s/pi * helper_inv<s>(order, nf, lm, alpha_s, cur_order + 1)
	  );
      }
    }
  }

  //!Decoupling constant for the mass in the \f$\overline{\rm MS}\f$ scheme.
  /**
   * @tparam s      Decoupling scheme
   * @param order   Perturbative order
   * @param nf      Number of active flavours before decoupling
   * @param lm      \f$\ln(\mu^2/m^2)\f$, where \f$\mu\f$ is the decoupling scale and \f$m\f$ the mass of the decoupled quark
   * @param alpha_s strong coupling constant at the decoupling scale
   * @returns       The decoupling constant to compute the mass in the theory with nf+1 flavours
   *
   * The decoupling constant \f$\zeta_m\f$ is defined by
   * \f$m^{(n_f-1)} = \zeta_m m^{(n_f)}\f$,
   * where \f$m^{(n_f)}\f$ is the mass in the theory with $n_f$ active flavours.
   *
   * All input parameters are defined in the theory with nf active flavours.
   */
  template<scheme s>
  inline QCD_CONSTEXPR_AFTER_CXX14
  double zeta_m(
      unsigned order,
      unsigned nf,
      double lm,
      double alpha_s
  ){
    using namespace detail::zeta_m;
    return helper<s>(order, nf, lm, alpha_s, 0);
  }

  //!Inverse decoupling constant for the mass in the \f$\overline{\rm MS}\f$ scheme.
  /**
   * @tparam s      Decoupling scheme
   * @param order   Perturbative order
   * @param nf      Number of active flavours before decoupling
   * @param lm      \f$\ln(\mu^2/m^2)\f$, where \f$\mu\f$ is the decoupling scale and \f$m\f$ the mass of the decoupled quark
   * @param alpha_s strong coupling constant with nf active flavours at the decoupling scale
   * @returns       The decoupling constant to compute the mass in the theory with nf+1 flavours
   *
   * The inverse decoupling constant \f$\zeta_m^{-1}\f$ is defined by
   * \f$m^{(n_f+1)} = \zeta_m^{-1} m^{(n_f)}\f$,
   * where \f$m^{(n_f)}\f$ is the mass in the theory with $n_f$ active flavours.
   *
   */
  template<scheme s>
  inline QCD_CONSTEXPR_AFTER_CXX14
  double zeta_m_inv(
      unsigned order,
      unsigned nf,
      double lm,
      double alpha_s
  ){
    using namespace detail::zeta_m;
    return helper_inv<s>(order, nf, lm, alpha_s, 0);
  }
  //!Coefficients of the decoupling constant.
  /**
   * @tparam s      Decoupling scheme
   * @param order   Perturbative order
   * @param nf      Number of active flavours before decoupling
   * @param lm      \f$\ln(\mu^2/m^2)\f$, where \f$\mu\f$ is the decoupling scale and \f$m\f$ the mass of the decoupled quark
   * @returns       The perturbative coefficient of the decoupling constant at the given order
   *
   * The coefficients \f$\zeta_{m, i}\f$ are defined via
   * \f$\zeta_m = \sum_{i=0}^\infty \zeta_{m, i} (\frac{\alpha_s}{\pi})^i\f$.
   */
  template<scheme s>
  inline QCD_CONSTEXPR_AFTER_CXX14
  double zeta_m(
      unsigned order,
      unsigned nf,
      double lm
  ){
    return detail::zeta_m::zeta_m<s>(order, nf, lm);
  }

  //!Coefficients of the inverse decoupling constant.
  /**
   * @tparam s      Decoupling scheme
   * @param order   Perturbative order
   * @param nf      Number of active flavours before decoupling
   * @param lm      \f$\ln(\mu^2/m^2)\f$, where \f$\mu\f$ is the decoupling scale and \f$m\f$ the mass of the decoupled quark
   * @returns       The perturbative coefficient of the inverse decoupling constant at the given order
   *
   * The coefficients \f$\zeta_{m, i}^{-1}\f$ are defined via
   * \f$\zeta_m^{-1} = \sum_{i=0}^\infty \zeta_{m, i}^{-1} (\frac{\alpha_s}{\pi})^i\f$.
   */
  template<scheme s>
  inline QCD_CONSTEXPR_AFTER_CXX14
  double zeta_m_inv(
      unsigned order,
      unsigned nf,
      double lm
  ){
    return detail::zeta_m::zeta_m_inv<s>(order, nf, lm);
  }

}

#undef QCD_CONSTEXPR_AFTER_CXX14
