/**
 * @file
 * @brief ---
 * @author  Andreas Maier <andreas.maier@durham.ac.uk>
 * @version 2.1.1
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2018 Andreas Maier
 *
 * This file is part of the QCD library.
 *
 * The QCD library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QCD library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QCD library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include "MSbar_mass.hpp"

#include <array>
#include <cmath>
#include <stdexcept>
#include <string>
#include <array>
#include <vector>

#include <boost/numeric/odeint.hpp>

#include "gamma_m.hpp"
#include "beta.hpp"
#include "array_2d.hpp"
#include "zeta_m.hpp"
#include "decoupling.hpp"

namespace QCD{

  namespace{
    class RGE {
    public:

      RGE(int nf, int order):
        beta_coeffs(order+1),
        gamma_m_coeffs(order+1)
      {
	for(int i = 0; i <= order; ++i){
	  beta_coeffs[i] = QCD::beta(i,nf);
	  gamma_m_coeffs[i] = QCD::gamma_m(i,nf);          
	}
      }

      using state_type = std::array<double, 2>;
      enum Idx {alpha_s, m_MS};
      
      void operator()(
		      state_type const & p,
		      state_type & d,  //d[XXX] = d/d(log(mu^2)) XXX
		      double const // log(mu^2)
		      ) const{
	d[alpha_s] = p[alpha_s] * beta(p[alpha_s]);
        d[m_MS] = p[m_MS] * gamma_m(p[alpha_s]);
      }
      
    private:
      static constexpr double pi = 3.141592653589793238462643383279;

      double beta(double alpha_s) const {
	assert(beta_coeffs.size()>0);
	const double as = alpha_s/pi;
	double result = 0;
	for(unsigned order = beta_coeffs.size()-1; order > 0;--order){
	  result += beta_coeffs[order];
	  result *= as;
	}
	return( -as*(result + beta_coeffs[0]));
      }

      double gamma_m(double alpha_s) const {
	assert(gamma_m_coeffs.size()>0);
	const double as = alpha_s/pi;
	double result = 0;
	for(unsigned order = gamma_m_coeffs.size()-1; order > 0;--order){
	  result += gamma_m_coeffs[order];
	  result *= as;
	}
	return( -as*(result + gamma_m_coeffs[0]));
      }
      
      std::vector<double> beta_coeffs, gamma_m_coeffs;
    };
  }
    
  MSbar_mass::MSbar_mass(
      reference_mass ref,
      strong_coupling alpha_s,
      int run_order
  ): m0_{ref.m}, as0_{alpha_s(ref.mu)}, mu0_{ref.mu},
     abserr_{default_error}, relerr_{default_error},
     nf_{ref.nf}, running_order_{run_order}
  {
    if(order() < 0 || order() > max_running_order){
      throw std::out_of_range(
          "order " + std::to_string(order()) + " not in range 0..."
          + std::to_string(max_running_order)
      );
    }
    if(alpha_s.nf() != ref.nf){
      throw std::invalid_argument(
          "number of flavours " + std::to_string(alpha_s.nf())
          + " in strong coupling does not match number of flavours "
          + std::to_string(nf()) + " in mass"
      );
    }
  }

  MSbar_mass::MSbar_mass(
      reference_mass ref,
      double alpha_s_mu0,
      int run_order
  ): m0_{ref.m}, as0_{alpha_s_mu0}, mu0_{ref.mu},
     abserr_{default_error}, relerr_{default_error},
     nf_{ref.nf}, running_order_{run_order}
  {
    if(order() < 0 || order() > max_running_order){
      throw std::out_of_range(
          "order " + std::to_string(order()) + " not in range 0..."
          + std::to_string(max_running_order)
      );
    }
  }

  double MSbar_mass::operator()(double mu) const{
    return evolve(mu)[RGE::m_MS];
  }

  Parameters MSbar_mass::with_alpha_s(double mu) const {
    const auto evol = evolve(mu);
    return {evol[RGE::alpha_s], evol[RGE::m_MS]};
  }

  int MSbar_mass::order() const{
    return running_order_;
  }

  int MSbar_mass::nf() const{
    return nf_;
  }

  double MSbar_mass::to_scale_invariant(double ep) const{
    double m = m0_;
    double mu = mu0_;
    while(std::abs(m-mu) > ep){
      mu = m;
      m = operator()(mu);
    }
    return m;
  }

  // TODO: fix (sometimes? loops for ever)
  // double MSbar_mass::to_scale_invariant() const{
  //   double m = m0;
  //   double mu = mu0;
  //   while(
  //       (m > mu && m > std::nextafter(mu,m))
  //       || (m < mu && m < std::nextafter(mu,m))
  //   ){
  //     mu = m;
  //     m = operator()(mu);
  //   }
  //   return m;
  // }

  std::array<double, 2> MSbar_mass::evolve(double mu) const {
    namespace odeint = boost::numeric::odeint;
    using stepper_type = odeint::runge_kutta_dopri5<RGE::state_type>;
    auto DEQ = RGE{nf(), order()};
    auto parameters = RGE::state_type{{as0_, m0_}};
    double const start = 2*log(mu0_);
    double const end = 2*log(mu);
    double const step_size = (end-start)/100.;
    odeint::integrate_adaptive(
        odeint::make_controlled<stepper_type>(abserr_, relerr_),
        DEQ, parameters, start, end, step_size
    );
    return parameters;
  }

  double
  MSbar_mass::decouple_mass_up(double alpha_s, threshold const & thr, int nf) const{
    switch(thr.dec_scheme){
    case scheme::pole:
      return zeta_m_inv<scheme::pole>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    case scheme::MSbar:
      return zeta_m_inv<scheme::MSbar>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    case scheme::scale_invariant:
      return zeta_m_inv<scheme::scale_invariant>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    default:
      throw std::invalid_argument("decoupling not implemented for scheme");
    }
  }

  double
  MSbar_mass::decouple_mass_down(double alpha_s, threshold const & thr, int nf) const{
    switch(thr.dec_scheme){
    case scheme::pole:
      return zeta_m<scheme::pole>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    case scheme::MSbar:
      return zeta_m<scheme::MSbar>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    case scheme::scale_invariant:
      return zeta_m<scheme::scale_invariant>(
	  order(), nf, 2*log(thr.mu/thr.m), alpha_s
      );
    default:
      throw std::invalid_argument("decoupling not implemented for scheme");
    }
  }

  MSbar_mass
  MSbar_mass::remove(threshold thr) const{
    return decouple_down(thr);
  }

  MSbar_mass
  MSbar_mass::decouple_down(threshold thr) const{
    MSbar_mass m_nfm1 = *this;
    const auto evol = evolve(thr.mu);
    const double m_nf_thr = evol[RGE::m_MS];
    const double alpha_s_nf_thr = evol[RGE::alpha_s];
    m_nfm1.nf_ = nf() - 1;
    m_nfm1.mu0_ = thr.mu;
    m_nfm1.m0_ = m_nf_thr*decouple_mass_down(alpha_s_nf_thr, thr, nf());
    m_nfm1.as0_ = decouple_alpha_s_down(alpha_s_nf_thr, thr, nf(), order());
    return m_nfm1;
  }

  MSbar_mass
  MSbar_mass::add(threshold thr) const{
    return decouple_up(thr);
  }

  MSbar_mass
  MSbar_mass::decouple_up(threshold thr) const{
    MSbar_mass m_nfp1 = *this;
    const auto evol = evolve(thr.mu);
    const double m_nf_thr = evol[RGE::m_MS];
    const double alpha_s_nf_thr = evol[RGE::alpha_s];
    m_nfp1.nf_ = nf() + 1;
    m_nfp1.mu0_ = thr.mu;
    m_nfp1.m0_ = m_nf_thr*decouple_mass_up(alpha_s_nf_thr, thr, nf());
    m_nfp1.as0_ = decouple_alpha_s_up(alpha_s_nf_thr, thr, nf(), order());
    return m_nfp1;
  }
  
  MSbar_mass & MSbar_mass::error(double err){
    abserr_ = err;
    relerr_ = err;
    return *this;
  }

  MSbar_mass & MSbar_mass::abs_error(double abs_err){
    abserr_ = abs_err;
    return *this;
  }

  MSbar_mass & MSbar_mass::rel_error(double rel_err){
    relerr_ = rel_err;
    return *this;
  }

}
