/**
 * @file
 * @brief ---
 * @author  Andreas Maier <andreas.maier@durham.ac.uk>
 * @version 2.1.1
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2018 Andreas Maier
 *
 * This file is part of the QCD library.
 *
 * The QCD library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QCD library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QCD library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include <cassert>
#include <cmath>
#include <vector>
#include <array>
#include <stdexcept>
#include <string>

#include <boost/numeric/odeint.hpp>

#include "QCD/alpha_s.hpp"
#include "QCD/alpha_s.hpp"
#include "QCD/beta.hpp"
#include "QCD/decoupling.hpp"

namespace QCD{

  strong_coupling::strong_coupling(
      reference_coupling ref,
      int order
  ): alpha_s_mu0(ref.alpha_s), mu0(ref.mu),
     abserr_{default_error}, relerr_{default_error},
     nf0{ref.nf}, running_order{order}
  {
    if(order < 0 || order > max_running_order){
      throw std::out_of_range(
          "order " + std::to_string(order) + " not in range 0..."
          + std::to_string(max_running_order)
      );
    }
  }

  double strong_coupling::operator()(double mu) const{
    return evolve(alpha_s_mu0, mu0, mu, nf0, abserr_, relerr_);
  }

  int strong_coupling::order() const{
    return running_order;
  }

  int strong_coupling::nf() const{
    return nf0;
  }

  namespace{
    typedef std::array<double,1> state_type;

    class strong_coupling_RGE{
      std::vector<double> beta_coeffs;

      double beta(double alpha_s) const{
	static constexpr double pi = 3.141592653589793238462643383279;
	assert(beta_coeffs.size()>0);
	const double as = alpha_s/pi;
	double result = 0;
	for(unsigned order = beta_coeffs.size()-1; order > 0;--order){
	  result += beta_coeffs[order];
	  result *= as;
	}
	return( -as*(result + beta_coeffs[0]));
      }

    public:
      strong_coupling_RGE(int nf, int order)
	: beta_coeffs(std::vector<double>(order+1))
      {
	for(int i = 0; i <= order; ++i){
	  beta_coeffs[i] = QCD::beta(i,nf);
	}
      }

      void operator()(
		      state_type const & alpha_s,
		      state_type & dalpha_s,  //dalpha_s[0] = d/d(log(mu^2)) alpha_s
		      double const // log(mu^2)
		      ) const{
	dalpha_s[0] = alpha_s[0] * beta(alpha_s[0]);
      }
    };


  }
  //evolve alpha_s(mu0) to mu (without decoupling)
  double
  strong_coupling::evolve(
      double alpha_s, double mu0, double mu, int nf,
      double abserr, double relerr
  ) const{
    namespace odeint = boost::numeric::odeint;
    using stepper_type = odeint::runge_kutta_dopri5<state_type>;
    auto RGE = strong_coupling_RGE{nf, order()};
    auto alpha_s_cur = state_type{{alpha_s}};
    double const start = 2*log(mu0);
    double const end = 2*log(mu);
    double const step_size = (end-start)/100.;
    odeint::integrate_adaptive(
        odeint::make_controlled<stepper_type>(abserr, relerr),
        RGE, alpha_s_cur, start, end, step_size
    );
    return alpha_s_cur[0];
  }

  strong_coupling
  strong_coupling::remove(threshold thr) const{
    return decouple_down(thr);
  };

  strong_coupling
  strong_coupling::decouple_down(threshold thr) const{
    strong_coupling const & alpha_s_nf = *this;
    strong_coupling alpha_s_nfm1 = alpha_s_nf;
    alpha_s_nfm1.mu0 = thr.mu;
    alpha_s_nfm1.alpha_s_mu0 = decouple_alpha_s_down(
        alpha_s_nf(thr.mu), thr, nf0, order()
    );
    --alpha_s_nfm1.nf0;
    return alpha_s_nfm1;
  }

  strong_coupling
  strong_coupling::add(threshold thr) const{
    return decouple_up(thr);
  };

  strong_coupling strong_coupling::decouple_up(threshold thr) const{
    strong_coupling const & alpha_s_nf = *this;
    strong_coupling alpha_s_nfp1 = alpha_s_nf;
    alpha_s_nfp1.mu0 = thr.mu;
    alpha_s_nfp1.alpha_s_mu0 = decouple_alpha_s_up(
        alpha_s_nf(thr.mu), thr, nf0, order()
    );
    ++alpha_s_nfp1.nf0;
    return alpha_s_nfp1;
  }

  strong_coupling & strong_coupling::error(double err){
    abserr_ = err;
    relerr_ = err;
    return *this;
  }

  strong_coupling & strong_coupling::abs_error(double abs_err){
    abserr_ = abs_err;
    return *this;
  }

  strong_coupling & strong_coupling::rel_error(double rel_err){
    relerr_ = rel_err;
    return *this;
  }

}
