/** test zeta_m
 * reference values from RunDec
 * note conventions
 *  zeta_alpha_s: nf = nl + 1
 *  zeta_alpha_s_inv: nf = nl
 **/

#include "QCD/zeta_m.hpp"

#include <iostream>
#include <cmath>
#include <cassert>
#include <array>
#include <stdexcept>
#include <string>

double diff(double tst, double ref){
  if(ref == 0.) return tst;
  return (tst - ref)/ref;
}

bool possible_zero(double d){
  static constexpr double ep = 1e-13;
  return std::abs(d) < ep;
}


template<class T, size_t... Rest> struct array;

template<class T, size_t N>
struct array<T, N>: std::array<T, N> {};

template<class T, size_t First, size_t Second, size_t... Rest>
struct array<T, First, Second, Rest...>:
  array<array<T, Second, Rest...>, First> {};

struct WrongResult: std::logic_error{
  WrongResult(
      std::string const & name,
      double my_value, double ref_value,
      int order, int nf_pow, int log_pow
  ):
    std::logic_error{
      "bad result in " + name + ": " + std::to_string(my_value) + " != "
      + std::to_string(ref_value)
      + " (order = " + std::to_string(order)
      + ", nf_pow = " + std::to_string(nf_pow)
      + ", log_pow = " + std::to_string(log_pow)
    } {}
};

#define TEST(FUN, SCHEME, REF) {                                         \
  const double my_value = QCD::FUN<scheme::SCHEME>(                     \
      order,nf_pow, log_pow                                              \
  );                                                    \
  const double ref_value = REF##_##SCHEME[order][nf_pow][log_pow]; \
  const double del = diff(my_value, ref_value); \
  std::cerr << #FUN #SCHEME << '(' << order << ", " << nf_pow << "): " << del << '\n'; \
  if(!possible_zero(del)){\
     throw WrongResult{#FUN #SCHEME, my_value, ref_value, order, nf_pow, log_pow}; \
  };                                                                    \
  }

int main(){
  using QCD::scheme;
  constexpr int max_order = QCD::zeta_m_max_order;

  // reference results ref[pert_order][nf_power][log_power]
  array<
    double,
    QCD::zeta_m_max_order + 1u, QCD::zeta_m_max_order - 1u, QCD::zeta_m_max_order + 1u
    > ref_OS, ref_MS, ref_SI, ref_inv_OS, ref_inv_MS, ref_inv_SI;


  ref_OS[0][0][0] = 1.;
  ref_OS[1][0][0] = 0.;
  ref_OS[1][0][1] = 0.;
  ref_OS[2][0][0] = 0.20601851851851852;
  ref_OS[2][0][1] = -0.1388888888888889;
  ref_OS[2][0][2] = 0.08333333333333333;
  ref_OS[3][0][0] = 1.4525287624862648;
  ref_OS[3][0][1] = -0.8323468020157112;
  ref_OS[3][0][2] = 0.7384259259259259;
  ref_OS[3][0][3] = 0.14351851851851852;
  ref_OS[3][1][0] = 0.02472760936815077;
  ref_OS[3][1][1] = -0.12268518518518519;
  ref_OS[3][1][2] = 0.;
  ref_OS[3][1][3] = -0.009259259259259259;
  ref_OS[4][0][0] = 1.783112700960487;
  ref_OS[4][0][1] = 3.325894126200277;
  ref_OS[4][0][2] = 7.176614089719983;
  ref_OS[4][0][3] = 2.328317901234568;
  ref_OS[4][0][4] = 0.30063657407407407;
  ref_OS[4][1][0] = -1.6159112400964275;
  ref_OS[4][1][1] = -0.6570014862863435;
  ref_OS[4][1][2] = -0.9738618827160493;
  ref_OS[4][1][3] = -0.16319444444444445;
  ref_OS[4][1][4] = -0.037037037037037035;
  ref_OS[4][2][0] = 0.05615582122474318;
  ref_OS[4][2][1] = 0.005015862378999758;
  ref_OS[4][2][2] = 0.023919753086419752;
  ref_OS[4][2][3] = 0.;
  ref_OS[4][2][4] = 0.0011574074074074073;

  ref_MS[0][0][0] = 1.;
  ref_MS[1][0][0] = 0.;
  ref_MS[1][0][1] = 0.;
  ref_MS[2][0][0] = 0.20601851851851852;
  ref_MS[2][0][1] = -0.1388888888888889;
  ref_MS[2][0][2] = 0.08333333333333333;
  ref_MS[3][0][0] = 1.822899132856635;
  ref_MS[3][0][1] = -0.9990134686823778;
  ref_MS[3][0][2] = 0.4050925925925926;
  ref_MS[3][0][3] = 0.14351851851851852;
  ref_MS[3][1][0] = 0.02472760936815077;
  ref_MS[3][1][1] = -0.12268518518518519;
  ref_MS[3][1][2] = 0.;
  ref_MS[3][1][3] = -0.009259259259259259;
  ref_MS[4][0][0] = 8.371928509821231;
  ref_MS[4][0][1] = -0.9588086344447362;
  ref_MS[4][0][2] = 1.8317066823125758;
  ref_MS[4][0][3] = 1.0088734567901234;
  ref_MS[4][0][4] = 0.30063657407407407;
  ref_MS[4][1][0] = -1.5780193327058316;
  ref_MS[4][1][1] = -0.1648174541674049;
  ref_MS[4][1][2] = -0.8025655864197531;
  ref_MS[4][1][3] = -0.0798611111111111;
  ref_MS[4][1][4] = -0.037037037037037035;
  ref_MS[4][2][0] = 0.05615582122474318;
  ref_MS[4][2][1] = 0.005015862378999758;
  ref_MS[4][2][2] = 0.023919753086419752;
  ref_MS[4][2][3] = 0.;
  ref_MS[4][2][4] = 0.0011574074074074073;

  ref_SI[0][0][0] = 1.;
  ref_SI[1][0][0] = 0.;
  ref_SI[1][0][1] = 0.;
  ref_SI[2][0][0] = 0.20601851851851852;
  ref_SI[2][0][1] = -0.1388888888888889;
  ref_SI[2][0][2] = 0.08333333333333333;
  ref_SI[3][0][0] = 1.822899132856635;
  ref_SI[3][0][1] = -1.2767912464601556;
  ref_SI[3][0][2] = 0.7384259259259259;
  ref_SI[3][0][3] = 0.14351851851851852;
  ref_SI[3][1][0] = 0.02472760936815077;
  ref_SI[3][1][1] = -0.12268518518518519;
  ref_SI[3][1][2] = 0.;
  ref_SI[3][1][3] = -0.009259259259259259;
  ref_SI[4][0][0] = 8.371928509821231;
  ref_SI[4][0][1] = -4.125817053290973;
  ref_SI[4][0][2] = 4.806243719349613;
  ref_SI[4][0][3] = 2.328317901234568;
  ref_SI[4][0][4] = 0.30063657407407407;
  ref_SI[4][1][0] = -1.5780193327058316;
  ref_SI[4][1][1] = -0.37160757762419505;
  ref_SI[4][1][2] = -0.8257137345679012;
  ref_SI[4][1][3] = -0.16319444444444445;
  ref_SI[4][1][4] = -0.037037037037037035;
  ref_SI[4][2][0] = 0.05615582122474318;
  ref_SI[4][2][1] = 0.005015862378999758;
  ref_SI[4][2][2] = 0.023919753086419752;
  ref_SI[4][2][3] = 0.;
  ref_SI[4][2][4] = 0.0011574074074074073;

  ref_inv_OS[0][0][0] = 1.;
  ref_inv_OS[1][0][0] = 0.;
  ref_inv_OS[1][0][1] = 0.;
  ref_inv_OS[2][0][0] = -0.20601851851851852;
  ref_inv_OS[2][0][1] = 0.1388888888888889;
  ref_inv_OS[2][0][2] = -0.08333333333333333;
  ref_inv_OS[3][0][0] = -1.4772563718544156;
  ref_inv_OS[3][0][1] = 0.8863591476947236;
  ref_inv_OS[3][0][2] = -0.6921296296296297;
  ref_inv_OS[3][0][3] = -0.16203703703703703;
  ref_inv_OS[3][1][0] = -0.02472760936815077;
  ref_inv_OS[3][1][1] = 0.12268518518518519;
  ref_inv_OS[3][1][2] = 0.;
  ref_inv_OS[3][1][3] = 0.009259259259259259;
  ref_inv_OS[4][0][0] = -0.30109112125205684;
  ref_inv_OS[4][0][1] = -3.7149415236110848;
  ref_inv_OS[4][0][2] = -5.541401336860276;
  ref_inv_OS[4][0][3] = -2.677854938271605;
  ref_inv_OS[4][0][4] = -0.33188657407407407;
  ref_inv_OS[4][1][0] = 1.5035995976469416;
  ref_inv_OS[4][1][1] = 0.6346059568442687;
  ref_inv_OS[4][1][2] = 0.9873649691358025;
  ref_inv_OS[4][1][3] = 0.16319444444444445;
  ref_inv_OS[4][1][4] = 0.03935185185185185;
  ref_inv_OS[4][2][0] = -0.05615582122474318;
  ref_inv_OS[4][2][1] = -0.005015862378999758;
  ref_inv_OS[4][2][2] = -0.023919753086419752;
  ref_inv_OS[4][2][3] = 0.;
  ref_inv_OS[4][2][4] = -0.0011574074074074073;

  ref_inv_MS[0][0][0] = 1.;
  ref_inv_MS[1][0][0] = 0.;
  ref_inv_MS[1][0][1] = 0.;
  ref_inv_MS[2][0][0] = -0.20601851851851852;
  ref_inv_MS[2][0][1] = 0.1388888888888889;
  ref_inv_MS[2][0][2] = -0.08333333333333333;
  ref_inv_MS[3][0][0] = -1.8476267422247858;
  ref_inv_MS[3][0][1] = 1.0530258143613902;
  ref_inv_MS[3][0][2] = -0.3587962962962963;
  ref_inv_MS[3][0][3] = -0.16203703703703703;
  ref_inv_MS[3][1][0] = -0.02472760936815077;
  ref_inv_MS[3][1][1] = 0.12268518518518519;
  ref_inv_MS[3][1][2] = 0.;
  ref_inv_MS[3][1][3] = 0.009259259259259259;
  ref_inv_MS[4][0][0] = -6.744671265486929;
  ref_inv_MS[4][0][1] = -0.09371909138130619;
  ref_inv_MS[4][0][2] = -0.30297541093434965;
  ref_inv_MS[4][0][3] = -1.2195216049382716;
  ref_inv_MS[4][0][4] = -0.33188657407407407;
  ref_inv_MS[4][1][0] = 1.4657076902563455;
  ref_inv_MS[4][1][1] = 0.14242192472532988;
  ref_inv_MS[4][1][2] = 0.8160686728395061;
  ref_inv_MS[4][1][3] = 0.0798611111111111;
  ref_inv_MS[4][1][4] = 0.03935185185185185;
  ref_inv_MS[4][2][0] = -0.05615582122474318;
  ref_inv_MS[4][2][1] = -0.005015862378999758;
  ref_inv_MS[4][2][2] = -0.023919753086419752;
  ref_inv_MS[4][2][3] = 0.;
  ref_inv_MS[4][2][4] = -0.0011574074074074073;

  ref_inv_SI[0][0][0] = 1.;
  ref_inv_SI[1][0][0] = 0.;
  ref_inv_SI[1][0][1] = 0.;
  ref_inv_SI[2][0][0] = -0.20601851851851852;
  ref_inv_SI[2][0][1] = 0.1388888888888889;
  ref_inv_SI[2][0][2] = -0.08333333333333333;
  ref_inv_SI[3][0][0] = -1.8476267422247858;
  ref_inv_SI[3][0][1] = 1.3308035921391679;
  ref_inv_SI[3][0][2] = -0.6921296296296297;
  ref_inv_SI[3][0][3] = -0.16203703703703703;
  ref_inv_SI[3][1][0] = -0.02472760936815077;
  ref_inv_SI[3][1][1] = 0.12268518518518519;
  ref_inv_SI[3][1][2] = 0.;
  ref_inv_SI[3][1][3] = 0.009259259259259259;
  ref_inv_SI[4][0][0] = -6.744671265486929;
  ref_inv_SI[4][0][1] = 3.142733771909377;
  ref_inv_SI[4][0][2] = -3.022882818341757;
  ref_inv_SI[4][0][3] = -2.677854938271605;
  ref_inv_SI[4][0][4] = -0.33188657407407407;
  ref_inv_SI[4][1][0] = 1.4657076902563455;
  ref_inv_SI[4][1][1] = 0.34921204818212004;
  ref_inv_SI[4][1][2] = 0.8392168209876543;
  ref_inv_SI[4][1][3] = 0.16319444444444445;
  ref_inv_SI[4][1][4] = 0.03935185185185185;
  ref_inv_SI[4][2][0] = -0.05615582122474318;
  ref_inv_SI[4][2][1] = -0.005015862378999758;
  ref_inv_SI[4][2][2] = -0.023919753086419752;
  ref_inv_SI[4][2][3] = 0.;
  ref_inv_SI[4][2][4] = -0.0011574074074074073;

  for(int order = 0; order <= max_order; ++order){
    const int max_nf_power = QCD::detail::zeta_m::max_nf_pow(order);
    for(int nf_pow = 0; nf_pow <= max_nf_power; ++nf_pow){
      const int max_log_power = QCD::detail::zeta_m::max_log_pow(order, nf_pow);
      for(int log_pow = 0; log_pow <= max_log_power; ++log_pow){
        TEST(zeta_m_log_coeff, OS, ref);
        TEST(zeta_m_log_coeff, MS, ref);
        TEST(zeta_m_log_coeff, SI, ref);
        TEST(zeta_m_inv_log_coeff, OS, ref_inv);
        TEST(zeta_m_inv_log_coeff, MS, ref_inv);
        TEST(zeta_m_inv_log_coeff, SI, ref_inv);
      }
    }
  }
}
