/**
 * @file
 * @brief Compile-time generation of arrays
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 2.2.1
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <array>

namespace array{
    /* generate a compile-time sequence of integers 0 ... N */
    template<size_t... Ns> struct sequence {}; /* the final sequence */
    /* recursion step */
    template<size_t N, size_t... Ns>
    struct gen_sequence : gen_sequence<N-1, N-1, Ns...> {};
    /* end of recursion */
    template<size_t... Ns>
    struct gen_sequence<0, Ns...> : sequence<Ns...> {};

  //! compile-time generation of array
  /**
   *  @tparam indices   Indices to be used as arguments for the generator.
   *  @param g          Function for generating array elements
   *  @returns          The array {g(i1), g(i2),...}, where i1,i2,... are
   *                    the indices indicated by the type of the second argument
   */
    template<class Generator, size_t... indices>
    auto constexpr
    gen_array_from_indices(Generator g, sequence<indices...>) ->
      std::array<decltype(g(std::size_t{})), sizeof...(indices)>{
      return {{ g(indices)... }};
    }
  //! compile-time generation of array from indices 0,...,N
  /**
   *  @tparam N   Highest index of the array
   *  @param g    Function for generating array elements
   *  @returns    The array {g(0), ..., g(N)}
   */
    template<size_t N, class Generator>
    auto constexpr
    gen_array(Generator g) -> std::array<decltype(g(std::size_t{})), N>{
      return gen_array_from_indices(g, gen_sequence<N>{});
    }
}
