/**
 * @file
 * @brief Physical constants
 *
 * Changes to the constants may require a recompilation of the whole
 * QQbar_threshold library to take effect. Wherever possible, it is
 * recommended to change settings via the
 * \ref QQbar_threshold::options "options"
 * parameter of the QQbar_threshold functions instead.
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 2.2.1
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once
#include <cmath>

/// -
namespace QQbar_threshold{

  //! Default value for strong coupling at the scale mZ
  constexpr double alpha_s_mZ = 0.1184;
  //! QED coupling at the scale mZ
  constexpr double alpha_mZ = 1/128.944;
  //! QED coupling at the scale mu_alpha_Y
  constexpr double alpha_Y = 1/132.274;
  //! Typical scale for Y resonances
  constexpr double mu_alpha_Y = 10.2;
  //! Mass of the Z boson
  constexpr double mZ = 91.1876;
  //! Mass of the W boson
  constexpr double mW = 80.385;
  //! Electron mass
  constexpr double me = 0.0005109989461;

  //! Fermi constant - this is exclusively used for calculating the top width
  constexpr double G_F = 0.000011663787;

  //! Default value for mass of the Higgs boson
  constexpr double m_Higgs = 125.;

  //! Fine structure constant
  constexpr double alpha_QED = 1/137.035999074;

  //! Electric charge of the top quark in units of the positron charge
  constexpr double e_u = 2./3.;
  //! Electric charge of the bottom quark in units of the positron charge
  constexpr double e_d = -1./3.;
  //! Electric charge of the electron in units of the positron charge
  constexpr double e = -1;
  //! cosine of the weak mixing angle squared
  constexpr double cw2 = mW*mW/(mZ*mZ);
  //! sine of the weak mixing angle squared
  constexpr double sw2 = 1 - cw2;

  //! Weak isospin of neutrino
  constexpr double T3_nu = 1./2.;
  //! Weak isospin of electron
  constexpr double T3_e = -1./2.;
  //! Weak isospin of top
  constexpr double T3_u = 1./2.;
  //! Weak isospin of bottom
  constexpr double T3_d = -1./2.;

  //! Reference scale-invariant mass for bottom quarks
  constexpr double mb_SI = 4.203;
  //! Decoupling threshold for bottom quarks
  constexpr double mu_thr = 2*mb_SI;

  //! Number of light flavours for bottom-related functions
  constexpr int nl_bottom = 4;
  //! Number of light flavours for top-related functions
  constexpr int nl_top = 5;

  //! Default PS scale for bottom
  constexpr double mu_f_bottom = 2.;

  //! Default PS scale for top
  constexpr double mu_f_top = 20.;

  //! Conversion factor from 1/GeV^2 to picobarn
  constexpr double invGeV2_to_pb = 389379300.;

  //! Axial vector coupling of fermion to Z boson
  /**
   * @param T3   Weak isospin
   * @returns    a_f, the axial vector coupling to the Z boson
   */
  inline double axial_Z_coupling(double T3, double){
    return T3/(2*sqrt(sw2*cw2));
  }

  //! Vector coupling of fermion to Z boson
  /**
   * @param T3   Weak isospin
   * @param e    Electric charge in units of positron charge
   * @returns    v_f, the vector coupling to the Z boson
   */
  inline double vector_Z_coupling(double T3, double e){
    return (T3 - 2*e*sw2)/(2*sqrt(sw2*cw2));
  }

  namespace detail{
    enum coupling{
      axial, vector,
      last_coupling = vector
    };

    //!\internal Z couplings to electron
    //TODO: make constexpr (requires constexpr sqrt)
    template<coupling c>
    double c_e();

    template<>
    inline double c_e<axial>(){
      static const double a_e = axial_Z_coupling(T3_e, e);
      return a_e;
    }

    template<>
    inline double c_e<vector>(){
      static const double v_e = vector_Z_coupling(T3_e, e);
      return v_e;
    }

    //!\internal Z couplings to neutrinos
    //TODO: make constexpr (requires constexpr sqrt)
    template<coupling c>
    double c_nu();

    template<>
    inline double c_nu<axial>(){
      static const double a_nu = axial_Z_coupling(T3_nu, 0);
      return a_nu;
    }

    template<>
    inline double c_nu<vector>(){
      static const double v_nu = vector_Z_coupling(T3_nu, 0);
      return v_nu;
    }
  }

}
