/**
 * @file
 * @brief Coefficients for double potential insertions
 *
 * This header contains the h functions required for the calculation
 * of double insertions with non-zero light quark masses.
 *
 * Reference:
 * Beneke, Maier, Piclum, Rauh, Nucl. Phys. B891 42-72, arXiv:1411.3132
 * \cite Beneke:2014pta
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 2.2.1
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once
#include "parameters.hpp"

namespace QQbar_threshold{
  namespace detail {

    //! h function for insertion of NLO potential with massive quark
    /**
     *  @tparam j    Order in the expansion lam -> n
     *  @param n     Principal quantum number
     *  @param s     Partial wave number
     *  @param xi_n  \f$ \frac{\alpha_s C_F m}{2 n m_{\rm light}} $\f
     *
     *  cf. eq. (C.20) in \cite Beneke:2014pta
     */
    template<int j>
    double h_light(int n, int s, double xi_n);

    //! h function for insertion of NLO potential with massless quarks
    /**
      * @tparam j    Order in the expansion lam -> n
      * @param n     Principal quantum number
      * @param s     Partial wave number
      * @param p     Further parameters
      *
      *  cf. eqs. (C.26)-(C.32) in \cite Beneke:2014pta
      */
    template<int j>
    double h(int n, int s, parameters const & p);

  }
}
