/**
 * @file
 * @brief Implementation of Green functions
 *
 * @see Green_functions.hpp
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 2.2.1
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include <complex>
#include <cmath>
#include <functional>

#include "complex.hpp"
#include "harmonic_sums.hpp"
#include "parameters.hpp"
#include "QCD/colour.hpp"
#include "zeta.hpp"
#include "psi.hpp"
#include "log_lam.hpp"
#include "insertions.hpp"
#include "Wilson_coefficients.hpp"
#include "Green_functions.hpp"
#include "integrate.hpp"
#include "Gamma.hpp"

namespace QQbar_threshold{
  namespace detail{
    using namespace complex;
    using QCD::c_F;
    using QCD::c_A;
    using QCD::beta;
    using zeta::pi;
    using harmonic_sums::harmonic_sum;
    using psi::psi;
    typedef std::complex<double> complex;


    /***************************************************/
    /*                LO Green function                */
    /***************************************************/

    template<>
    complex Green_function<0>(complex lam, parameters const & p){
      return -pow(p.m,2)*c_F*p.alpha_s/(4*pi)*(
	  + 1/(2*lam)
	  + log_lam(lam, p)
	  - 1./2.
	  + harmonic_sum<1>(- lam)
      );
    }

    template<>
    complex Green_function_PR_mu_weak_in_real<0>(
	complex lam, parameters const & p
    ){
      const double log_w = log(p.mu/p.mu_w);
      return
	+ Green_function_PR<0>(lam,p)
	- pow(p.m,2)*c_F*p.alpha_s/(4*pi)*log_w
	;
    }

    double Green_function_pole(parameters const & p){
      return pow(p.m,2)*p.alpha_s*c_F/(16.*pi);
    }

    /***************************************************/
    /*                NLO Green function               */
    /***************************************************/

    namespace{
      //! Insertion of QED Coulomb potential
      complex delta_G1_QED(complex lam, parameters const & p);
      //! Insertion of massive light-quark correction Coulomb potential
      complex delta_G1_light(complex lam, parameters const & p);
    }

    /**
     * we also multiply the QED Coulomb contribution with
     * the factor contributions.v_Coulomb[0]
     * Formally, this is wrong, but would have to be changed
     * also in a1_eff
     */
    template<>
    complex Green_function<1>(complex lam, parameters const & p){
      using Wilson_coefficients::a;
      using Wilson_coefficients::a_ep;
      return
        p.contributions.v_Coulomb[0]*(
            + p.contributions.v_QED_Coulomb[0]*delta_G1_QED(lam, p)
            + p.alpha_s/(4*pi)*(4*pi*c_F*p.alpha_s)*(
                + a_ep<1>(p.nl) * pow(p.m,2)/(96*pow(pi,2))
                + a<1>(p.nl) * insertion_1<1>(lam, p)
                + 4*beta(0, p.nl) * insertion<1+ep>(lam, p)
                + delta_G1_light(lam, p)
            )
        );
    }

    template<>
    complex Green_function_PR_mu_weak_in_real<1>(
	complex lam, parameters const & p
    ){
      using Wilson_coefficients::a;
      const double log_w = log(p.mu/p.mu_w);
      return
	+ Green_function_PR<1>(lam,p)
	+ log_w*p.m*p.m*p.contributions.v_Coulomb[0]*(
	    + c_F/(16*pi*pi)*p.alpha_s*p.alpha_s*(
		+ 4*beta(0, p.nl)*(1 - log_w)
		- a<1>(p.nl)
	    )
            - p.contributions.v_QED_Coulomb[0]*p.alpha/(4*pi)*p.e_q*p.e_q
	);
    }


    /***************************************************/
    /*               NNLO Green function               */
    /***************************************************/

    namespace{
      //! Insertion of NNLO Coulomb potential
      complex delta_G2_Coulomb(complex lam, parameters const & p);
      //! Double insertion of NLO Coulomb potential
      complex delta_G2_double_Coulomb(complex lam, parameters const & p);
      //! Insertion of delta (1/m^2) potential
      complex delta_G2_delta(complex lam, parameters const & p);
      //! Insertion of 1/r^2 (1/m) potential
      complex delta_G2_r2inv(complex lam, parameters const & p);
      //! Insertion of correction to kinetic energy
      complex delta_G2_kinetic(complex lam, parameters const & p);
      //! Insertion of momentum-dependent potential
      complex delta_G2_p2(complex lam, parameters const & p);
      //! Width^2 insertion
      complex delta_G2_width2(complex lam, parameters const & p);
      //! Width*p^2 insertion
      complex delta_G2_width_kinetic(complex lam, parameters const & p);
    }

    template<>
    complex Green_function<2>(complex lam, parameters const & p){
      return
	+ p.contributions.v_Coulomb[1] * delta_G2_Coulomb(lam, p)
	+ pow(p.contributions.v_Coulomb[0], 2) * delta_G2_double_Coulomb(lam, p)
	+ p.contributions.v_delta[0] * delta_G2_delta(lam, p)
	+ p.contributions.v_r2inv[0] * delta_G2_r2inv(lam, p)
	+ p.contributions.v_kinetic[0] * delta_G2_kinetic(lam, p)
	+ p.contributions.v_p2[0] * delta_G2_p2(lam, p)
        + p.contributions.v_width2[0] * delta_G2_width2(lam, p)
        + p.contributions.v_width_kinetic[0] * delta_G2_width_kinetic(lam, p)
	;
    }


    /***************************************************/
    /*               NNNLO Green function               */
    /***************************************************/

    namespace{
      //! Insertion of NNNLO Coulomb potential
      complex delta_G3_Coulomb(complex lam, parameters const & p);
      //! Double insertion of NNLO and NLO Coulomb potential
      complex delta_G3_double_Coulomb(complex lam, parameters const & p);
      //! Triple insertion of NLO Coulomb potential
      complex delta_G3_triple_Coulomb(complex lam, parameters const & p);
      //! Double insertion of NLO Coulomb and delta (1/m^2) potential
      complex delta_G3_Coulomb_delta(complex lam, parameters const & p);
      //! Double insertion of NLO Coulomb and 1/r^2 (1/m) potential
      complex delta_G3_Coulomb_r2inv(complex lam, parameters const & p);
      //! Double insertion of NLO Coulomb potential and kinetic energy correction
      complex delta_G3_Coulomb_kinetic(complex lam, parameters const & p);
      //! Double insertion of NLO Coulomb and momentum-dependent potential
      complex delta_G3_Coulomb_p2(complex lam, parameters const & p);
      //! Insertion of correction to delta (1/m^2) potential
      complex delta_G3_delta(complex lam, parameters const & p);
      //! Insertion of correction to 1/r^2 (1/m) potential
      complex delta_G3_r2inv(complex lam, parameters const & p);
      //! Insertion of correction to momentum-dependent potential
      complex delta_G3_p2(complex lam, parameters const & p);
      //! Ultrasoft correction
      complex delta_G3_ultrasoft(complex lam, parameters const & p);
      //! Insertion of Higgs potential
      complex delta_G3_Higgs(complex lam, parameters const & p);
    }

    template<>
    complex Green_function<3>(complex lam, parameters const & p){
      return
	+ p.contributions.v_Coulomb[2] * delta_G3_Coulomb(lam, p)
	+ p.contributions.v_Coulomb[1]*p.contributions.v_Coulomb[0] * delta_G3_double_Coulomb(lam, p)
	+ pow(p.contributions.v_Coulomb[0], 3) * delta_G3_triple_Coulomb(lam, p)
	+ p.contributions.v_delta[0]*p.contributions.v_Coulomb[0] * delta_G3_Coulomb_delta(lam, p)
	+ p.contributions.v_r2inv[0]*p.contributions.v_Coulomb[0] * delta_G3_Coulomb_r2inv(lam, p)
	+ p.contributions.v_p2[0]*p.contributions.v_Coulomb[0] * delta_G3_Coulomb_p2(lam, p)
	+ p.contributions.v_kinetic[0]*p.contributions.v_Coulomb[0] * delta_G3_Coulomb_kinetic(lam, p)
	+ p.contributions.v_delta[1] * delta_G3_delta(lam, p)
	+ p.contributions.v_r2inv[1] * delta_G3_r2inv(lam, p)
	+ p.contributions.v_p2[1] * delta_G3_p2(lam, p)
	+ p.contributions.ultrasoft[0] * delta_G3_ultrasoft(lam, p)
	+ p.contributions.v_Higgs[0] * delta_G3_Higgs(lam, p)
	;
    }

    /**************** counterterms from weak correction *****************
     *                                                                  *
     * These are corrections from choosing the scale mu_weak associated *
     * with finite-width divergences different from the overall         *
     * renormalisation scale. Consequently, they are all proportional   *
     * to log(mu/mu_weak).                                              *
     *                                                                  *
     ********************************************************************/

    template<>
    double Green_function_weak<0>(parameters const & p){
      const double log_w = log(p.mu/p.mu_w);
      return -4*log_w*pow(p.alpha_s/(4*pi),2)*(
	  + p.contributions.v_kinetic[0]
	  + p.contributions.v_p2[0]
	  - 1./3.*p.contributions.dv[0]
      );
    }

    template<>
    double Green_function_weak<1>(parameters const & p){
      if(p.mu == p.mu_w) return 0;
      /* parametrise scheme dependent finite term of NNLO GreenFunction */
      static constexpr double flag_fin = 1;
      const double log_w = log(p.mu/p.mu_w);
      using namespace Wilson_coefficients;
      double result =
	+ p.contributions.v_p2[1]*(
	    -4*4*beta(0, p.nl) + 32./3.*c_A
	)
	+ p.contributions.ultrasoft[0]*(
	    -8./3.*c_A - 16./3.*c_F
	)
	+ p.contributions.v_Coulomb[0]*(
	    - p.contributions.v_kinetic[0]*4*4*beta(0, p.nl)
	    + p.contributions.dv[0]*4./3.*4*beta(0, p.nl)
	)
	+ p.contributions.dv[1]*128./3.*c_F
	;
      result *= log_w;
      result +=
	+ p.contributions.v_delta[1]*(
	    -88./3.*c_A + 112./3.*c_F + 16*4*beta(0, p.nl)*v0(p.nl)
	)
	+ p.contributions.v_p2[1]*(
	    -4*vp1(p.nl) + 4*4*beta(0, p.nl)*flag_fin
	)
	+ p.contributions.ultrasoft[0]*(
	  + 152./3.*c_A - 16./3.*c_F
	  - 48*c_A*log(2) + 32*c_F*log(2)
	  - (16*c_A + 32*c_F)*log(p.mu/p.m)
	)
	+ p.contributions.cv[0]*cv_QCD<1>(p)*(
	    - p.contributions.v_kinetic[0]*8
	    - p.contributions.v_p2[0]*8
	    + p.contributions.dv[0]*4./3.
	)
	+ p.contributions.v_Coulomb[0]*(
	    + p.contributions.v_kinetic[0]*(
		- 4*a1_eff(p) + 4*flag_fin*4*beta(0, p.nl)
	    )
	    + p.contributions.dv[0]*(
		4./3.*a1_eff(p) - 4./3.*flag_fin*4*beta(0, p.nl)
	    )
	)
	+ p.contributions.dv[1]*(
	    4./3.*dv_QCD<1>(p) - 64./3.*flag_fin*c_F
	)
	;
      return pow(p.alpha_s/(4*pi),3)*log_w*result;
    }

    namespace{
      using Gamma::Gamma;

      complex delta_G1_QED(complex lam, parameters const & p){
        return 4*pi*pow(p.e_q, 2)*p.alpha*insertion<1>(lam, p);
      }

      /************************* NLO light-quark correction *********************
       *                                                                        *
       * NLO massive light-quark correction to the Green function.              *
       * Equation numbers refer to                                              *
       * Beneke, Maier, Piclum, Rauh, Nucl. Phys. B891 42-72, arXiv:1411.3132	*
       * \cite Beneke:2014pta							*
       *									*
       **************************************************************************/

      //! Integrand for part A of NLO light-quark correction
      /**
       * See eq. (D.3).
       * Part A is the part with zero extra gluons exchanged.
       */
      complex integrand_0_ex(complex u, complex xi){
	return u*(1-u)*Gamma(1./2.-u)*pow(Gamma(-u)*Gamma(u)*Gamma(u+1./2.),2)
	  /(Gamma(1+2*u)*Gamma(5./2.-u))*pow(xi/2,-2*u);
      }

      double re_integrand_0_ex(double y, complex xi){
	static constexpr double del = 0.25;
	auto u = complex{del, log(y/(1-y))};
	return real(integrand_0_ex(u, xi))/(y*(1-y));
      }

      double im_integrand_0_ex(double y, complex xi){
	static constexpr double del = 0.25;
	auto u = complex{del, log(y/(1-y))};
	return imag(integrand_0_ex(u, xi))/(y*(1-y));
      }

      //! Part A of NLO light-quark correction to Green function
      /**
       * As defined by eq. (D.3), up to a prefactor.
       * Part A is the part with zero extra gluons exchanged.
       *
       * The integral is performed over y with
       * u = del + i*log(y/(1-y))
       * where 0 < del < 1
       */
      complex delta_G1_light_0_ex(complex xi){
	using namespace std::placeholders;
	auto re_integrand = std::bind(re_integrand_0_ex, _1, xi);
	auto re = make_integrator(re_integrand);
	auto im_integrand = std::bind(im_integrand_0_ex, _1, xi);
	auto im = make_integrator(im_integrand);
	return complex{re.integrate(0,1), im.integrate(0,1)}/(8*pi*pi);
      }

      //! Inner integrand for part B of NLO light-quark correction
      /**
       * See last line of eq. (D.5).
       * Part B is the part with at least one extra gluon exchanged.
       */
      complex inner_all_ex(complex u, complex w, complex xi){
	return pow(xi,-u)*Gamma(2+u)*Gamma(-u)*Gamma(2-u/2)*Gamma(1+u+w)
	  /(u*Gamma((5-u)/2)*Gamma(1+u-w));
      }

      double re_inner_all_ex(double y, complex w, complex xi){
	static constexpr double del = 0.5;
	auto u = complex{del, log(y/(1-y))};
	return real(inner_all_ex(u, w, xi))/(y*(1-y));
      }

      double im_inner_all_ex(double y, complex w, complex xi){
	static constexpr double del = 0.5;
	auto u = complex{del, log(y/(1-y))};
	return imag(inner_all_ex(u, w, xi))/(y*(1-y));
      }

      //! Outer integrand for part B of NLO light-quark correction
      /**
       * See last two lines of eq. (D.5).
       * Part B is the part with at least one extra gluon exchanged.
       */
      complex integrand_all_ex(complex w, complex lam, complex xi){
	using namespace std::placeholders;
	auto re_integrand = std::bind(re_inner_all_ex, _1, w, xi);
	auto re_inner = make_integrator(re_integrand);
	auto im_integrand = std::bind(im_inner_all_ex, _1, w, xi);
	auto im_inner = make_integrator(im_integrand);
	return Gamma(w+1-lam)*Gamma(-w-1)*pow(Gamma(-w),2)
	  *complex{re_inner.integrate(0,1), im_inner.integrate(0,1)};
      }

      double re_integrand_all_ex(double y, complex lam, complex xi){
	const double del = (real(lam)-1)/2;
	assert(del < 0);
	auto u = complex{del, log(y/(1-y))};
	return real(integrand_all_ex(u, lam, xi))/(y*(1-y));
      }

      double im_integrand_all_ex(double y, complex lam, complex xi){
	const double del = (real(lam)-1)/2;
	assert(del < 0);
	auto u = complex{del, log(y/(1-y))};
	return imag(integrand_all_ex(u, lam, xi))/(y*(1-y));
      }

      //! Part B of NLO light-quark correction
      /**
       * Defined as in eq. (D.5), up to a prefactor.
       * Part B is the part with at least one extra gluon exchanged.
       */
      complex delta_G1_light_all_ex(complex lam, complex xi){
	using namespace std::placeholders;
	auto re_integrand = std::bind(re_integrand_all_ex, _1, lam, xi);
	auto re = make_integrator(re_integrand);
	auto im_integrand = std::bind(im_integrand_all_ex, _1, lam, xi);
	auto im = make_integrator(im_integrand);
	return -sqrt(pi)/Gamma(-lam)*complex{re.integrate(0,1), im.integrate(0,1)}/(4*pi*pi);
      }

      //! NLO light-quark correction to Green function
      /**
       * See eq. (D.2),
       * \delta_1 G_{m_c} there corresponds to
       * alpha_s/(4*Pi)*(4*pi*c_F*p.alpha_s)*delta_G1_light
       */
      complex delta_G1_light(complex lam, parameters const & p){
	if(p.ml == 0) return 0;
	complex const xi = sqrt(- p.E_shifted*p.m)/p.ml;
	return pow(p.m/(4*pi),2)*(
	    delta_G1_light_0_ex(xi) + delta_G1_light_all_ex(lam, xi)
	);
      }

      complex delta_G2_Coulomb(complex lam, parameters const & p){
	using Wilson_coefficients::a;
	return (4*pi*c_F*p.alpha_s)*pow(p.alpha_s/(4*pi),2)*(
    	    + a<2>(p.nl)*insertion<1>(lam, p)
    	    + (
		+ 2*a<1>(p.nl)*4*beta(0, p.nl)
		+ 16*beta(1, p.nl)
	    )*insertion<1+ep>(lam, p)
    	    + pow(4*beta(0, p.nl),2)*insertion<1+2*ep>(lam, p)
    	);
      }

      complex delta_G2_double_Coulomb(complex lam, parameters const & p){
	using Wilson_coefficients::a1_eff;
	return pow(c_F,2)*pow(p.alpha_s,4)*(
    	    + pow(a1_eff(p),2)*insertion<1, 1>(lam, p)
    	    + 2*a1_eff(p)*4*beta(0, p.nl)*insertion<1, 1+ep>(lam, p)
    	    + pow(4*beta(0, p.nl),2)*insertion<1+ep, 1+ep>(lam, p)
    	);
      }

      complex delta_G2_delta(complex lam, parameters const & p){
	using Wilson_coefficients::v0;
	using Wilson_coefficients::v0_ep;
	return c_F*p.alpha_s * (
    	    + 4/pow(p.m,2)*pi*v0(p.nl)*insertion<0>(lam, p)
    	    + 1./2.*c_F*p.alpha_s*v0_ep(p.nl)*Green_function<0>(lam, p)
    	);
      }

      complex delta_G2_r2inv(complex lam, parameters const & p){
      using Wilson_coefficients::b;
      using Wilson_coefficients::b_ep;
      return c_F*pow(p.alpha_s,2)*(
	  - pow(pi,2)/p.m*b<1>(p.nl)*insertion_R1<one_half+ep>(lam, p)
	  - 1/4.*b_ep<1>(p.nl)*Green_function<0>(lam, p)
    	);
      }

      complex delta_G2_kinetic(complex lam, parameters const & p){
	return pow(c_F*p.alpha_s,2)*(1./4.)*(
    	    + p.m*pow(c_F*p.alpha_s,2)/pow(2*lam,4)*insertion<delta>(lam, p)
    	    - Green_function<0>(lam, p)/(2*pow(lam,2))
    	    - (2*c_F*pi*p.alpha_s)/pow(lam,2)*insertion<1>(lam, p)
    	    + 2*pow(pi,2)/p.m*insertion_R1<one_half+ep>(lam, p)
    	    + log(2)*Green_function<0>(lam, p)
    	);
      }

      complex delta_G2_p2(complex lam, parameters const & p){
	return (
    	    4*pi*pow(c_F*p.alpha_s,2)*(
    		- c_F*p.alpha_s*insertion<1>(lam, p)/(4*pow(lam,2))
    		+ 1/(2*p.m)*pi*insertion_R1<one_half+ep>(lam, p)
    		+ log(2)*Green_function<0>(lam, p)/(4*pi)
    	    )
    	);
      }

      //! cf. eq. (36) in arXiv:1711.10429
      complex delta_G2_width2(complex lam, parameters const & p){
        return -p.width*p.width/(16*pi*c_F*p.alpha_s)*(
            lam*(1 + 2*lam*(1 + lam*psi<1>(1 - lam)))
        );
      }

      //! cf. eq. (37) in arXiv:1711.10429
      complex delta_G2_width_kinetic(complex lam, parameters const & p){
        return complex{0, -c_F/(8*pi)*p.m*p.alpha_s*p.width}*(
            - 3./4./lam + 1./2. - 2.*harmonic_sum<1>(-lam)
            + lam/2.*psi<1>(1-lam) - 2.*log_lam(lam, p.mu_w, p)
        );
      }

      using namespace Wilson_coefficients;

      complex delta_G3_Coulomb(complex lam, parameters const & p){
	return c_F*pow(p.alpha_s,4)/(16*pi*pi)*(
	    + a<3>(p.nl)*insertion<1>(lam, p)
    	   + (
	       + 8*pow(pi,2)*pow(c_A, 3)
	       + 64*beta(2, p.nl)
	       + 3*a<2>(p.nl)*4*beta(0, p.nl)
	       + 2*a<1>(p.nl)*16*beta(1, p.nl)
	   )*insertion<1+ep>(lam, p)
    	   + (
	       + 5./2.*4*beta(0, p.nl)*16*beta(1, p.nl)
	       + 3*a<1>(p.nl)*pow(4*beta(0, p.nl), 2)
	   )*insertion<1+2*ep>(lam, p)
    	   + pow(4*beta(0, p.nl), 3)*insertion<1+3*ep>(lam, p)
    	);
      }

      complex delta_G3_double_Coulomb(complex lam, parameters const & p){
	return 2*pow(4*pi*c_F*p.alpha_s,2)*pow(p.alpha_s/(4*pi),3)*(
    	    + a1_eff(p)*a<2>(p.nl)*insertion<1, 1>(lam, p)
    	    + (
    	    	2*a1_eff(p)*a<1>(p.nl)*4*beta(0, p.nl)
	    	+ a<2>(p.nl)*4*beta(0, p.nl)
	    	+ a1_eff(p)*16*beta(1, p.nl)
    	    )*insertion<1, 1+ep>(lam, p)
    	    + a1_eff(p)*pow(4*beta(0, p.nl), 2)*insertion<1, 1+2*ep>(lam, p)
    	    + (
                2*a<1>(p.nl)*pow(4*beta(0, p.nl), 2)
    	       + 4*beta(0, p.nl)*16*beta(1, p.nl)
    	       )*insertion<1+ep, 1+ep>(lam, p)
    	    + pow(4*beta(0, p.nl), 3)*insertion<1+ep, 1+2*ep>(lam, p)
    	);
      }

      complex delta_G3_triple_Coulomb(complex lam, parameters const & p){
	return pow(c_F, 3)*pow(p.alpha_s, 6)*(
      	    + pow(a1_eff(p),3)*insertion<1, 1, 1>(lam, p)
      	    + pow(a1_eff(p),2)*4*beta(0, p.nl)*(
      		+2*insertion<1, 1, 1+ep>(lam, p)
      		+ insertion<1, 1+ep, 1>(lam, p)
      	    )
      	    + a1_eff(p)*pow(4*beta(0, p.nl), 2)*(
      		+ 2*insertion<1, 1+ep, 1+ep>(lam, p)
      		+ insertion<1+ep, 1, 1+ep>(lam, p)
      	    )
      	    + pow(4*beta(0, p.nl), 3)*insertion<1+ep, 1+ep, 1+ep>(lam, p)
      	);
      }

      complex delta_G3_Coulomb_delta(complex lam, parameters const & p){
	return pow(c_F,2)*pow(p.alpha_s,3)*(
	    + 8*v0(p.nl)*pi/pow(p.m,2)*Green_function<0>(lam, p)*(
	     	+ a<1>(p.nl)*insertion_1<1>(lam, p)
                + (a1_eff(p) - a<1>(p.nl))*insertion<1>(lam, p)
    	    	+ 4*beta(0, p.nl)*insertion<1+ep>(lam, p)
	    )
	    + 1./2.*v0_ep(p.nl)*c_F*p.alpha_s*(
    	    	+ a1_eff(p)*insertion<1>(lam, p)
    	    	+ 4*beta(0, p.nl)*insertion<1+ep>(lam, p)
	    )
    	    + (
	    	+ v0_ep(p.nl)*(a1_eff(p) - 4*beta(0, p.nl))
	    	+ v0(p.nl)*a_ep<1>(p.nl)
	    	- 1./2.*v0_ep2(p.nl)*4*beta(0, p.nl)
	    )*Green_function<0>(lam, p)/(12*pi)
    	);
      }

      complex delta_G3_Coulomb_r2inv(complex lam, parameters const & p){
	return -pow(c_F/2.,2)*pow(p.alpha_s,4)*(
    	    + b_ep<1>(p.nl)*(
    		+ a1_eff(p)*insertion<1>(lam, p)
    		+ 4*beta(0, p.nl)*insertion<1+ep>(lam, p)
    	    )
    	    + b<1>(p.nl)*(
    		+ a1_eff(p)*insertion<1, one_half>(lam, p)
    		+ 4*beta(0, p.nl)*insertion<1+ep, one_half>(lam, p)
    	    )
    	);
      }

      complex delta_G3_Coulomb_p2(complex lam, parameters const & p){
	return 2*4*pi*pow(c_F,2)*pow(p.alpha_s,3)*(
	    - pow(c_F*p.alpha_s,2)/(4*pow(lam,2))*(
    	    	+ 4*beta(0, p.nl)*insertion<1, 1+ep>(lam, p)
    	    	+ a1_eff(p)*insertion<1, 1>(lam, p)
    	    )
    	    + c_F*p.alpha_s/(16*pi)*(
    	    	+ a1_eff(p)*insertion<1, one_half>(lam, p)
    	    	+ 4*beta(0, p.nl)*insertion<1+ep, one_half>(lam, p)
    	    )
    	    + c_F*p.alpha_s*log(2)/(8*pi)*(
    	    	+ a1_eff(p)*insertion<1>(lam, p)
    	    	+ 4*beta(0, p.nl)*insertion<1+ep>(lam, p)
    	    )
    	    + 1/(16*p.m)*(
    	    	+ 4*beta(0, p.nl)*(
    	    	    + insertion_R0<one_half+2*ep>(lam, p)
    	    	    - insertion_R0<one_half+ep>(lam, p)
    	    	)
    	    	+ (a1_eff(p) + 4*beta(0, p.nl)*log(4))
	    	*insertion_R1<one_half+2*ep>(lam, p)
    	    	- 4*beta(0, p.nl)*log(4)*insertion_R1<one_half+ep>(lam, p)
    	    	+ (
    	    	    + a_ep<1>(p.nl)
    	    	    + 17./12.*pow(pi,2)*4*beta(0, p.nl)
    	    	    + (a1_eff(p) + 4*beta(0, p.nl)*log(2))*log(4)
    	    	)*insertion_R2<one_half+2*ep>(lam, p)
    	    	- 1./12.*4*beta(0, p.nl)*(
    	    	    5*pow(pi,2) + 6*pow(log(4),2)
    	    	)*insertion_R2<one_half+ep>(lam, p)
	    )
    	);
      }

      complex delta_G3_Coulomb_kinetic(complex lam, parameters const & p){
	return -((1./2.)*pow(c_F,2)*pow(p.alpha_s,3)*(
		     - pow(c_F, 3)*p.m*pow(p.alpha_s, 3)/(16*pow(lam,4))*(
			 + 4*beta(0, p.nl)*insertion<1+ep, delta>(lam, p)
			 + a1_eff(p)*insertion<1, delta>(lam, p)
		     )
		     + c_F*p.alpha_s/(2*pow(lam,2))*(
			 + 4*beta(0, p.nl)*insertion<1+ep>(lam, p)
			 + a<1>(p.nl)*insertion_1<1>(lam, p)
			 + (a1_eff(p)-a<1>(p.nl))*insertion<1>(lam, p)
			 + a_ep<1>(p.nl)*(pow(p.m,2)/(96*pow(pi,2)))
		     )
		     - pi/(2*p.m)*(
			 + 4*beta(0, p.nl)*(
			     + insertion_R0<one_half+2*ep>(lam, p)
			     - insertion_R0<one_half+ep>(lam, p)
			 )
			 + (
			     a1_eff(p) + 4*beta(0, p.nl)*log(4)
			 )*insertion_R1<one_half+2*ep>(lam, p)
			 - 4*beta(0, p.nl)*log(4)*insertion_R1<one_half+ep>(lam, p)
			 + (
			     + a_ep<1>(p.nl)
			     + 17./12.*pow(pi,2)*4*beta(0, p.nl)
			     + (a1_eff(p) + 4*beta(0, p.nl)*log(2))*log(4)
			 )*insertion_R2<one_half+2*ep>(lam, p)
			 - 1./12.*4*beta(0, p.nl)*(
			     + 5*pow(pi,2) + 6*pow(log(4),2)
			 )*insertion_R2<one_half+ep>(lam, p)
		     )
		     + 2*pow(c_F, 2)*pi*pow(p.alpha_s, 2)/pow(lam,2)*(
			 + 4*beta(0, p.nl)*insertion<1, 1+ep>(lam, p)
			 + a1_eff(p)*insertion<1, 1>(lam, p)
		     )
		     - 1/4.*c_F*p.alpha_s*(
			 + a1_eff(p)*insertion<1, one_half>(lam, p)
			 + 4*beta(0, p.nl)*insertion<1+ep, one_half>(lam, p)
		     )
		     - 1/4.*c_F*p.alpha_s*log(4)*(
			 + a1_eff(p)*insertion<1>(lam, p)
			 + 4*beta(0, p.nl)*insertion<1+ep>(lam, p)
		     )
		 ));
      }
      complex delta_G3_delta(complex lam, parameters const & p){
	return - c_F*pow(p.alpha_s/p.m,2)*(
    	    - (
	    	+ 7./3.*c_F
	    	- 11./6.*c_A
	    	+ v0(p.nl)*4*beta(0, p.nl)
	    )*insertion_R0<ep>(lam, p)
	    - (vq1(p.nl) + v0_ep(p.nl)*4*beta(0, p.nl))*insertion_R1<ep>(lam, p)
	    - (vq1_ep(p.nl) + v0_ep2(p.nl)*4*beta(0, p.nl))*insertion_R2<ep>(lam, p)
    	    + pow(p.m,2)*c_F*p.alpha_s/(8*pi)*(
    	    	+ v0_ep2(p.nl)*4*beta(0, p.nl) - vm1_ep(p.nl) - vm1(p.nl)*2*log(p.mu/p.m)
    	    	- (c_A + 2./3.*c_F)*pow(log(p.mu/p.m),2)
    	    )*Green_function<0>(lam, p)
    	    + (
                -vm1(p.nl) + v0_ep(p.nl)*4*beta(0, p.nl) - (c_A + 2./3.*c_F)*log(p.mu/p.m)
    	    )*insertion<0>(lam, p)
    	);
      }

      complex delta_G3_r2inv(complex lam, parameters const & p){
	return -c_F*pi*pow(p.alpha_s, 3)/p.m*(
    	    + (-2./3.*pow(c_A, 2) - 4./3.*c_A*c_F)*insertion_R0<one_half+2*ep>(lam, p)
    	    + b<2>(p.nl)*insertion_R1<one_half+2*ep>(lam, p)
    	    + b_ep<2>(p.nl)*insertion_R2<one_half+2*ep>(lam, p)
    	    + b<1>(p.nl)/2    *4*beta(0, p.nl)*(
		+ insertion_R0<one_half+2*ep>(lam, p)
		- insertion_R0<one_half+ep>(lam, p)
	    )
    	    + b_ep<1>(p.nl)/2 *4*beta(0, p.nl)*(
		+ insertion_R1<one_half+2*ep>(lam, p)
		- insertion_R1<one_half+ep>(lam, p)
	    )
    	    + b_ep2<1>(p.nl)/2*4*beta(0, p.nl)*(
		+ insertion_R2<one_half+2*ep>(lam, p)
		- insertion_R2<one_half+ep>(lam, p)
	    )
    	);
      }

      complex delta_G3_p2(complex lam, parameters const & p){
	return -(
    	    + pow(c_F, 3)*pow(p.alpha_s, 4)/(4*pow(lam,2))*(
    		+ vp1_ep(p.nl)*pow(p.m,2)/(96*pow(pi,2))
    		+ vp1(p.nl)*insertion_1<1>(lam, p)
    		+ (8./3.*c_A + 4*beta(0, p.nl))*insertion<1+ep>(lam, p)
    	    )
    	    + pow(c_F, 2)*pi*pow(p.alpha_s, 3)*(
    		- 1/(2*p.m)*(8./3.*c_A + 4*beta(0, p.nl))*(
    		    + insertion_R0<one_half+2*ep>(lam, p)
		    - insertion_R0<one_half+ep>(lam, p)
    		)
    		- (
		    + 2*log(2)*(8./3.*c_A + 4*beta(0, p.nl))
		    + vp1(p.nl)
		)/(2*p.m)*insertion_R1<one_half+2*ep>(lam, p)
    		+ (
		    + 8./3.*c_A
		    + 4*beta(0, p.nl)
		)*log(4)/(2*p.m)*insertion_R1<one_half+ep>(lam, p)
    		- (
    		   vp1_ep(p.nl) + 2*log(2)*vp1(p.nl)
    		   + (17./12.*pow(pi,2) + 2*pow(log(2),2))*(8./3.*c_A + 4*beta(0, p.nl))
    		)/(2*p.m)*insertion_R2<one_half+2*ep>(lam, p)
    		+ (8./3.*c_A + 4*beta(0, p.nl))
    		*(5*pow(pi,2) + 24*pow(log(2),2))/(24*p.m)
    		*insertion_R2<one_half+ep>(lam, p)
    	    )
    	);
      }

      complex delta_G3_Higgs(complex lam, parameters const & p){
	return
	  p.Yukawa_t2/(2*pow(p.m_Higgs,2))
	  *insertion<0>(lam, p)
	  ;
      }

      // ultrasoft correction
      // TODO: use simpler formula from arXiv:0804.4004
      inline complex delta_G3_ultrasoft(complex lam, parameters const & p){
	return
	  pow(p.m,2)*pow(p.alpha_s,4)/(3888*pow(lam,3)*pow(pi,2))*(
	      3888*pow(lam,3)*pow(pi,2)*insertion_US(lam)
	      + c_F*(
		  1536*log(p.alpha_s) - 6368*lam*log(p.alpha_s) - 42592*pow(lam,2)*log(p.alpha_s)
		  + 3648*lam*log(2)*log(p.alpha_s) + 9792*pow(lam,2)*log(2)*log(p.alpha_s)
		  + 6528*lam*log(c_F/lam)*log(p.alpha_s) + 57120*pow(lam,2)*log(c_F/lam)*
		  log(p.alpha_s) - 93932*pow(lam,3)*log(c_F/lam)*log(p.alpha_s)
		  + 19584*pow(lam,3)*log(2)*log(c_F/lam)*log(p.alpha_s)
		  + 57120*pow(lam,3)*pow(log(c_F/lam),2)*log(p.alpha_s) + 3264*lam*pow(log(p.alpha_s),2)
		  + 28560*pow(lam,2)*pow(log(p.alpha_s),2) - 57120*gamma_Euler*pow(lam,3)*pow(log(p.alpha_s),2)
		  + 28560*pow(lam,3)*(-2)*log(lam)*pow(log(p.alpha_s),2) + 768*log(p.m/p.mu)
		  - 3912*lam*log(p.m/p.mu) - 19512*pow(lam,2)*log(p.m/p.mu)
		  + 2736*lam*log(2)*log(p.m/p.mu) + 3672*pow(lam,2)*log(2)*log(p.m/p.mu)
		  + 3264*lam*log(c_F/lam)*log(p.m/p.mu) + 35904*pow(lam,2)*log(c_F/lam)*
		  log(p.m/p.mu) - 45585*pow(lam,3)*log(c_F/lam)*log(p.m/p.mu)
		  + 7344*pow(lam,3)*log(2)*log(c_F/lam)*log(p.m/p.mu)
		  + 35904*pow(lam,3)*pow(log(c_F/lam),2)*log(p.m/p.mu) + 3264*lam*log(p.alpha_s)*
		  log(p.m/p.mu) + 35904*pow(lam,2)*log(p.alpha_s)*log(p.m/p.mu)
		  - 71808*gamma_Euler*pow(lam,3)*log(p.alpha_s)*log(p.m/p.mu)
		  + 35904*pow(lam,3)*(-2)*log(lam)*log(p.alpha_s)*log(p.m/p.mu)
		  + 408*lam*pow(log(p.m/p.mu),2) + 9996*pow(lam,2)*pow(log(p.m/p.mu),2)
		  - 19992*gamma_Euler*pow(lam,3)*pow(log(p.m/p.mu),2) + 9996*pow(lam,3)*(-2)*log(lam)*
		  pow(log(p.m/p.mu),2) + 1632*pow(lam,3)*pow(harmonic_sum<1>(-lam),2)*
		  (35*log(p.alpha_s) + 22*log(p.m/p.mu)) + lam*harmonic_sum<1>(-lam)*
		  (
		      57120*pow(lam,2)*pow(log(p.alpha_s),2)
		      + 3*log(p.m/p.mu)*(
			  1088 + 11968*lam
			  - 15195*pow(lam,2) + 2448*pow(lam,2)*log(2) + 23936*pow(lam,2)*
			  log(c_F/lam) + 6664*pow(lam,2)*log(p.m/p.mu)
		      )
		      + 4*log(p.alpha_s)*(
			  1632 + 14280*lam - 23483*pow(lam,2)
			  + 4896*pow(lam,2)*log(2) + 28560*pow(lam,2)*log(c_F/lam)
			  + 17952*pow(lam,2)*log(p.m/p.mu))
		  )
		  + 27*pow(lam,2)*(
		      + 4*(- 32 + 81*pow(lam,2))*log(p.alpha_s)
		      + (- 64 + 243*pow(lam,2))*log(p.m/p.mu)
		  )*psi<1>(1-lam)
	      )
	  );
      }

    }

  }
}
