/**
 * @file
 * @brief Set parameters for cross section and bound state functions
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 2.2.1
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2016, 2017 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include "parameters.hpp"
#include <array>

#include "scheme_conversion.hpp"
#include "alpha_s.hpp"
#include "zeta.hpp"

#include "not_implemented.hpp"

namespace QQbar_threshold{
  using zeta::pi;

  namespace pert_order{
    std::string to_string(order o){
      switch(o){
      case LO: return "LO";
      case NLO: return "NLO";
      case N2LO: return "N2LO";
      case N3LO: return "N3LO";
      default:;
      }
      return "(unknown order)";
    }

    std::ostream& operator<<(std::ostream& os, order o){
      os << to_string(o);
      return os;
    }
  }

  std::string to_string(SM s){
    switch(s){
    case SM::none: return "none";
    case SM::all: return "all";
    case SM::QED: return "QED";
    case SM::Higgs: return "Higgs";
    case SM::QEDHiggs: return "QED | Higgs";
    default:
      throw std::exception{}; //unreachable
    }
  }

  std::ostream& operator<<(std::ostream& os, SM s){
    os << to_string(s);
    return os;
  }

  std::string to_string(production_channel p){
    switch(p){
    case production_channel::photon_only: return "photon_only";
    case production_channel::S_wave_only: return "S_wave_only";
    case production_channel::all: return "all";
    }
    throw std::exception{}; //unreachable
  }

  std::ostream& operator<<(std::ostream& os, production_channel p){
    os << to_string(p);
    return os;
  }

  std::string to_string(alpha_s_options o){
    using std::to_string;
    return "{" + to_string(o.mu0) + ", " + to_string(o.alpha_s) + "}";
  }

  namespace{
    //! Generate default bottom options
    options gen_bottom_opts(){
      options opt;
      opt.alpha = alpha_Y;
      opt.mu_alpha = mu_alpha_Y;
      opt.expand_s = false;
      opt.mass_scheme = {PSshift, mu_f_bottom, mu_f_bottom};
      opt.beyond_QCD = SM::QED;
      opt.resonant_only = true;
      opt.invariant_mass_cut = 0;
      return opt;
    }

    //! Generate default top options
    options gen_top_opts(){
      options opt;
      opt.alpha = alpha_mZ;
      opt.mu_alpha = mZ;
      opt.expand_s = true;
      opt.mass_scheme = {PSshift, mu_f_top, mu_f_top};
      opt.beyond_QCD = SM::all;
      opt.resonant_only = false;
      return opt;
    }
  };

  options bottom_options(){
    static const options result = gen_bottom_opts();
    return result;
  }

  options top_options(){
    static const options result = gen_top_opts();
    return result;
  }

  namespace{
    template<size_t N>
      std::string to_string(std::array<double, N> a){
      using std::to_string;
      std::string result = to_string(a[0]);
      for(size_t i = 1; i < a.size(); ++i){
	result += ", " + to_string(a[i]);
      }
      return result;
    }

    template<typename T>
      std::string contributions_to_string(T const & opt){
      return
	"contributions.v_Coulomb = {" + to_string(opt.contributions.v_Coulomb) + "};\n"
	"contributions.v_delta = {" + to_string(opt.contributions.v_delta) + "};\n"
	"contributions.v_r2inv = {" + to_string(opt.contributions.v_r2inv) + "};\n"
	"contributions.v_p2 = {" + to_string(opt.contributions.v_p2) + "};\n"
	"contributions.v_kinetic = {" + to_string(opt.contributions.v_kinetic) + "};\n"
	"contributions.ultrasoft = {" + to_string(opt.contributions.ultrasoft) + "};\n"
	"contributions.v_QED_Coulomb = {" + to_string(opt.contributions.v_QED_Coulomb) + "};\n"
	"contributions.v_width2 = {" + to_string(opt.contributions.v_width2) + "};\n"
	"contributions.v_width_kinetic = {" + to_string(opt.contributions.v_width_kinetic) + "};\n"
	"contributions.width_ep = {" + to_string(opt.contributions.width_ep) + "};\n"
	"contributions.v_Higgs = {" + to_string(opt.contributions.v_Higgs) + "};\n"
	"contributions.cv = {" + to_string(opt.contributions.cv) + "};\n"
	"contributions.cv_Higgs = {" + to_string(opt.contributions.cv_Higgs) + "};\n"
	"contributions.Cv_QED = {" + to_string(opt.contributions.Cv_QED) + "};\n"
	"contributions.Cv_WZ = {" + to_string(opt.contributions.Cv_WZ) + "};\n"
	"contributions.Ca_QED = {" + to_string(opt.contributions.Ca_QED) + "};\n"
	"contributions.Ca_WZ = {" + to_string(opt.contributions.Ca_WZ) + "};\n"
	"contributions.dv = {" + to_string(opt.contributions.dv) + "};\n"
	"contributions.ca = {" + to_string(opt.contributions.ca) + "};\n"
        "contributions.nonresonant = {" + to_string(opt.contributions.nonresonant) + "};\n"
	;
    }
  }

  std::string to_string(options const & opt){
    using std::to_string;
    using QQbar_threshold::to_string;
    return
      contributions_to_string(opt)
      + "alpha_s_mZ = " + to_string(opt.alpha_s_mZ) + ";\n"
      "alpha_s_mu0 = " + to_string(opt.alpha_s_mu0) + ";\n"
      "m_Higgs = " + to_string(opt.m_Higgs) + ";\n"
      "Yukawa_factor = " + to_string(opt.Yukawa_factor) + ";\n"
      "invariant_mass_cut = " + to_string(opt.invariant_mass_cut) + ";\n"
      "ml = " + to_string(opt.ml) + ";\n"
      "r4 = " + to_string(opt.r4) + ";\n"
      "alpha = " + to_string(opt.alpha) + ";\n"
      "mu_alpha = " + to_string(opt.mu_alpha) + ";\n"
      "resum_poles = " + to_string(opt.resum_poles) + ";\n"
      "beyond_QCD = " + to_string(opt.beyond_QCD) + ";\n"
      "mass_scheme = " + to_string(opt.mass_scheme) + ";\n"
      "production = " + to_string(opt.production) + ";\n"
      "resonant_only = " + (opt.resonant_only?"true":"false") + ";\n"
      "expand_s = " + (opt.expand_s?"true":"false") + ";\n"
      "double_light_insertion = "
      + (opt.double_light_insertion?"true":"false")
      + ";\n"
      "ISR_const = " + (opt.ISR_const?"true":"false")
      ;
  }

  std::ostream& operator<<(std::ostream& os, options const & opt){
    os << to_string(opt);
    return os;
  }

  std::string to_string(parameters const & p){
    using std::to_string;
    return
      contributions_to_string(p)
      + "E_shifted = {"
      + to_string(real(p.E_shifted)) + ", " + to_string(imag(p.E_shifted))
      + "};\n"
      "s = " + to_string(p.s) + ";\n"
      "m = " + to_string(p.m) + ";\n"
      "width = " + to_string(p.width) + ";\n"
      "mu = " + to_string(p.mu) + ";\n"
      "alpha_s = " + to_string(p.alpha_s) + ";\n"
      "m_Higgs = " + to_string(p.m_Higgs) + ";\n"
      "Yukawa_t2 = " + to_string(p.Yukawa_t2) + ";\n"
      "invariant_mass_cut = " + to_string(p.invariant_mass_cut) + ";\n"
      "ml = " + to_string(p.ml) + ";\n"
      "e_q = " + to_string(p.e_q) + ";\n"
      "a_q = " + to_string(p.c_q[detail::axial]) + ";\n"
      "v_q = " + to_string(p.c_q[detail::vector]) + ";\n"
      "r4 = " + to_string(p.r4) + ";\n"
      "mu_w = " + to_string(p.mu_w) + ";\n"
      "alpha = " + to_string(p.alpha) + ";\n"
      "mu_alpha = " + to_string(p.mu_alpha) + ";\n"
      "order = " + to_string(p.order) + ";\n"
      "nl = " + to_string(p.nl) + ";\n"
      "resum_poles = " + to_string(p.resum_poles) + ";\n"
      "beyond_QCD = " + to_string(p.beyond_QCD) + ";\n"
      "mass_scheme = " + to_string(p.mass_scheme) + ";\n"
      "production = " + to_string(p.production) + ";\n"
      "resonant_only = " + (p.resonant_only?"true":"false") + ";\n"
      "expand_s = " + (p.expand_s?"true":"false") + ";\n"
      "double_light_insertion = "
      + (p.double_light_insertion?"true":"false")
      + ";\n"
      "ISR_const = " + (p.ISR_const?"true":"false") + ";\n"
      ;
  }

  std::string bottom_internal_settings(
      double mu, double m, order o, options opt
  ){
    return to_string(gen_bottom_parameters(mu, m, o, opt));
  }
  std::string bottom_internal_settings(
      double sqrt_s, double mu, double m, order o, options opt
  ){
    return to_string(gen_bottom_parameters(sqrt_s, mu, m, o, opt));
  }
  std::string top_internal_settings(double mu, double m, order o, options opt){
    return to_string(gen_top_parameters(mu, m, o, opt));
  }
  std::string top_internal_settings(
      double sqrt_s, double mu, top_properties M, order o, options opt
  ){
    return to_string(gen_top_parameters(sqrt_s, mu, M, o, opt));
  }
  std::string top_internal_settings(
      double sqrt_s, scales mu, top_properties M, order o, options opt
  ){
    return to_string(gen_top_parameters(sqrt_s, mu, M, o, opt));
  }

  namespace {
    //! Constant for MSbar/pole conversion
    /**
     * results from Marquard, Smirnov, Smirnov, Steinhauser, Wellmann
     * Phys.Rev. D94 (2016) no.7, 074025, arXiv:1606.06754
     * Eqs. (21) and (38)
     *
     * Note that our MSbar/pole conversion is expressed in terms of alpha_s(nl),
     * so r4 = c_m^{(4)} + \delta c_m^{(4)} with mu = m(mu)
     */
    double get_default_r4(int nl){
      return 3562.20 + nl*(-745.290 + (43.3963 - 0.678141*nl)*nl);
    }

    //! Set parameters for all but the xsection functions
    /*
     *  p.alpha_s and p.nl have to be set before calling this function!
     */
    void set_parameters(
	parameters & p,
	double mu, double m, double width, order o, options opt
    ){
      p.mu = mu;
      p.m = m;
      p.width = width;
      p.order = o;
      p.contributions = std::move(opt.contributions);
      p.m_Higgs = opt.m_Higgs;
      p.invariant_mass_cut = opt.invariant_mass_cut;
      p.ml = opt.ml;
      p.alpha = opt.alpha;
      p.mu_alpha = opt.mu_alpha;
      p.beyond_QCD = opt.beyond_QCD;
      p.mass_scheme = opt.mass_scheme;
      /*
         In the manual we state that opt.mass_scheme should be defined by
         what is known as aggregate initialization. If only one scale is given,
         the remaining scale is guarenteed to be value-initialized to 0., which
         tells us to use the default nu = scale
      */
      if(p.mass_scheme.nu == 0.) p.mass_scheme.nu = p.mass_scheme.scale;
      p.production = opt.production;
      p.resum_poles = opt.resum_poles;
      p.expand_s = opt.expand_s;
      p.double_light_insertion = opt.double_light_insertion;
      p.ISR_const = opt.ISR_const;

      //apply correction settings
      if(! (p.beyond_QCD & SM::QED)){
	std::fill(
	    begin(p.contributions.v_QED_Coulomb),
	    end(p.contributions.v_QED_Coulomb),
	    0.0
	);
	std::fill(
	    begin(p.contributions.Cv_QED),
	    end(p.contributions.Cv_QED),
	    0.0
	);
	std::fill(
	    begin(p.contributions.Ca_QED),
	    end(p.contributions.Ca_QED),
	    0.0
	);
      }
      if(! (p.beyond_QCD & SM::Higgs)){
	std::fill(
	    begin(p.contributions.v_Higgs),
	    end(p.contributions.v_Higgs),
	    0.0
	);
	std::fill(
	    begin(p.contributions.cv_Higgs),
	    end(p.contributions.cv_Higgs),
	    0.0
	);
      }
      if(p.beyond_QCD != SM::all){
	std::fill(
	    begin(p.contributions.Cv_WZ),
	    end(p.contributions.Cv_WZ),
	    0.0
	);
	std::fill(
	    begin(p.contributions.Ca_WZ),
	    end(p.contributions.Ca_WZ),
	    0.0
	);
	std::fill(
	    begin(p.contributions.width_ep),
	    end(p.contributions.width_ep),
	    0.0
	);
	std::fill(
	    begin(p.contributions.v_width2),
	    end(p.contributions.v_width2),
	    0.0
	);
	std::fill(
	    begin(p.contributions.v_width_kinetic),
	    end(p.contributions.v_width_kinetic),
	    0.0
	);
      }

      if((o > NNLO) && (p.mass_scheme.name == MSbar) && std::isnan(opt.r4)){
	p.r4 = get_default_r4(p.nl);
      }
      else{
	p.r4 = opt.r4;
      }
      /*
       * Set inital value for Yukawa factor
       * this is needed for shift schemes if the scheme conversion
       * depends on the Yukawa coupling (e.g. 1S-shift)
       */
      p.Yukawa_t2 = 2*pi/(sw2*cw2)*p.alpha*pow(opt.Yukawa_factor*p.m/mZ,2);
      // treat shifted masses (zero for Pole, PS insertion,...)
      p.m += detail::mass_shift(p);
      // final value for Yukawa factor
      p.Yukawa_t2 = 2*pi/(sw2*cw2)*p.alpha*pow(opt.Yukawa_factor*p.m/mZ,2);
    }

    //! Set parameters for xsection functions
    /*
     *  p.alpha_s and p.nl have to be set before calling this function!
     */
    void set_parameters(
	parameters & p,
	double sqrt_s, scales mu, double m, double width, order o,
	options opt
    ){
      set_parameters(p, mu.soft, m, width, o, opt);
      p.resonant_only = opt.resonant_only;
      p.mu_w = mu.weak;
      if(std::isnan(opt.invariant_mass_cut)){
	p.invariant_mass_cut = p.m - mW;
      }
      else{
	p.invariant_mass_cut = opt.invariant_mass_cut;
      }
      p.s = sqrt_s * sqrt_s;
      p.E_shifted = complex{sqrt_s - 2 * p.m, width};
      if(is_threshold_mass(p.mass_scheme)){
	p.E_shifted -= 2*detail::delta_m(0,p);
      }
    }

    parameters gen_bottom_parameters(double mu, options const & opt){
      parameters p;
      p.nl = nl_bottom;
      p.e_q = e_d;
      p.c_q[detail::axial]  = axial_Z_coupling(T3_d, e_d);
      p.c_q[detail::vector] = vector_Z_coupling(T3_d, e_d);
      p.alpha_s = std::isnan(opt.alpha_s_mu0.mu0)?
	alpha_s_bottom(mu, opt.alpha_s_mZ):
	alpha_s_bottom(mu, opt.alpha_s_mu0);
      return p;
    }

    parameters gen_top_parameters(double mu, options const & opt){
      parameters p;
      p.nl = nl_top;
      p.e_q = e_u;
      p.c_q[detail::axial]  = axial_Z_coupling(T3_u, e_u);
      p.c_q[detail::vector] = vector_Z_coupling(T3_u, e_u);
      p.alpha_s = std::isnan(opt.alpha_s_mu0.mu0)?
	alpha_s_top(mu, opt.alpha_s_mZ):
	alpha_s_top(mu, opt.alpha_s_mu0);
      return p;
    }

  }

  parameters gen_bottom_parameters(
      double mu, double m, order o,
      options opt
  ){
    parameters p = gen_bottom_parameters(mu, opt);
    set_parameters(p, mu, m, 0, o, opt);
    return p;
  }

  parameters gen_bottom_parameters(
      double sqrt_s, double mu, double m, order o,
      options opt
  ){
    return gen_bottom_parameters(sqrt_s, {mu, mu}, m, o, opt);
  }

  parameters gen_bottom_parameters(
      double sqrt_s, scales mu, double m, order o,
      options opt
  ){
    static constexpr double width = 1e-9;
    parameters p = gen_bottom_parameters(mu.soft, opt);
    set_parameters(p, sqrt_s, mu, m, width, o, opt);
    return p;
  }

  parameters gen_top_parameters(
      double mu, double m, order o,
      options opt
  ){
    return gen_top_parameters(mu, {m, 0.}, o, opt);
  }
  parameters gen_top_parameters(
      double mu, top_properties M, order o,
      options opt
  ){
    parameters p = gen_top_parameters(mu, opt);
    set_parameters(p, mu, M.m, M.width, o, opt);
    p.width = M.width;
    return p;
  }
  parameters gen_top_parameters(
      double sqrt_s, double mu, top_properties M, order o,
      options opt
  ){
    return gen_top_parameters(sqrt_s, {mu, mu}, M, o, opt);
  }
  parameters gen_top_parameters(
      double sqrt_s, scales mu, top_properties M, order o,
      options opt
  ){
    parameters p = gen_top_parameters(mu.soft, opt);
    set_parameters(p, sqrt_s, mu, M.m, M.width, o, opt);
    return p;
  }


}
